# Copyright 2010-2011 Josh Durgin
# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Storage backend for RBD
   (RADOS (Reliable Autonomic Distributed Object Store) Block Device)"""

import contextlib
import logging
import math
import urllib

from eventlet import tpool
from oslo_config import cfg
from oslo_utils import eventletutils
from oslo_utils import units

from glance_store import capabilities
from glance_store.common import utils
from glance_store import driver
from glance_store import exceptions
from glance_store.i18n import _, _LE, _LI, _LW
from glance_store import location

try:
    import rados
    import rbd
except ImportError:
    rados = None
    rbd = None

DEFAULT_POOL = 'images'
DEFAULT_USER = None    # let librados decide based on the Ceph conf file
DEFAULT_CHUNKSIZE = 8  # in MiB
DEFAULT_SNAPNAME = 'snap'

LOG = logging.getLogger(__name__)

_RBD_OPTS = [
    cfg.IntOpt('rbd_store_chunk_size', default=DEFAULT_CHUNKSIZE,
               min=1,
               help="""
Size, in megabytes, to chunk RADOS images into.

Provide an integer value representing the size in megabytes to chunk
Glance images into. The default chunk size is 8 megabytes. For optimal
performance, the value should be a power of two.

When Ceph's RBD object storage system is used as the storage backend
for storing Glance images, the images are chunked into objects of the
size set using this option. These chunked objects are then stored
across the distributed block data store to use for Glance.

Possible Values:
    * Any positive integer value

Related options:
    * None

"""),
    cfg.StrOpt('rbd_store_pool', default=DEFAULT_POOL,
               help="""
RADOS pool in which images are stored.

When RBD is used as the storage backend for storing Glance images, the
images are stored by means of logical grouping of the objects (chunks
of images) into a ``pool``. Each pool is defined with the number of
placement groups it can contain. The default pool that is used is
'images'.

More information on the RBD storage backend can be found here:
http://ceph.com/planet/how-data-is-stored-in-ceph-cluster/

Possible Values:
    * A valid pool name

Related options:
    * None

"""),
    cfg.StrOpt('rbd_store_user', default=DEFAULT_USER,
               help="""
RADOS user to authenticate as.

This configuration option takes in the RADOS user to authenticate as.
This is only needed when RADOS authentication is enabled and is
applicable only if the user is using Cephx authentication. If the
value for this option is not set by the user or is set to None, a
default value will be chosen, which will be based on the client.
section in rbd_store_ceph_conf.

Possible Values:
    * A valid RADOS user

Related options:
    * rbd_store_ceph_conf

"""),
    cfg.StrOpt('rbd_store_ceph_conf', default='',
               help="""
Ceph configuration file path.

This configuration option specifies the path to the Ceph configuration
file to be used. If the value for this option is not set by the user
or is set to the empty string, librados will read the standard ceph.conf
file by searching the default Ceph configuration file locations in
sequential order.  See the Ceph documentation for details.

NOTE: If using Cephx authentication, this file should include a reference
to the right keyring in a client.<USER> section

NOTE 2: If you leave this option empty (the default), the actual Ceph
configuration file used may change depending on what version of librados
is being used.  If it is important for you to know exactly which configuration
file is in effect, you may specify that file here using this option.

Possible Values:
    * A valid path to a configuration file

Related options:
    * rbd_store_user

"""),
    cfg.IntOpt('rados_connect_timeout', default=-1,
               help="""
Timeout value for connecting to Ceph cluster.

This configuration option takes in the timeout value in seconds used
when connecting to the Ceph cluster i.e. it sets the time to wait for
glance-api before closing the connection. This prevents glance-api
hangups during the connection to RBD. If the value for this option
is set to less than 0, no timeout is set and the default librados value
is used.

Possible Values:
    * Any integer value

Related options:
    * None

"""),
    cfg.BoolOpt('rbd_thin_provisioning',
                default=False,
                help="""
Enable or not thin provisioning in this backend.

This configuration option enable the feature of not really write null byte
sequences on the RBD backend, the holes who can appear will automatically
be interpreted by Ceph as null bytes, and do not really consume your storage.
Enabling this feature will also speed up image upload and save network traffic
in addition to save space in the backend, as null bytes sequences are not
sent over the network.

Possible Values:
    * True
    * False

Related options:
    * None

"""),
]


class StoreLocation(location.StoreLocation):
    """
    Class describing a RBD URI. This is of the form:

        rbd://image

        or

        rbd://fsid/pool/image/snapshot
    """

    def process_specs(self):
        # convert to ascii since librbd doesn't handle unicode
        for key, value in self.specs.items():
            self.specs[key] = str(value)
        self.fsid = self.specs.get('fsid')
        self.pool = self.specs.get('pool')
        self.image = self.specs.get('image')
        self.snapshot = self.specs.get('snapshot')

    def get_uri(self):
        if self.fsid and self.pool and self.snapshot:
            # ensure nothing contains / or any other url-unsafe character
            safe_fsid = urllib.parse.quote(self.fsid, '')
            safe_pool = urllib.parse.quote(self.pool, '')
            safe_image = urllib.parse.quote(self.image, '')
            safe_snapshot = urllib.parse.quote(self.snapshot, '')
            return "rbd://%s/%s/%s/%s" % (safe_fsid, safe_pool,
                                          safe_image, safe_snapshot)
        else:
            return "rbd://%s" % self.image

    def parse_uri(self, uri):
        prefix = 'rbd://'
        self.validate_schemas(uri, valid_schemas=(prefix,))
        # convert to ascii since librbd doesn't handle unicode
        try:
            ascii_uri = str(uri)
        except UnicodeError:
            reason = _('URI contains non-ascii characters')
            msg = _LI("Invalid URI: %s") % reason

            LOG.info(msg)
            raise exceptions.BadStoreUri(message=reason)
        pieces = ascii_uri[len(prefix):].split('/')
        if len(pieces) == 1:
            self.fsid, self.pool, self.image, self.snapshot = \
                (None, None, pieces[0], None)
        elif len(pieces) == 4:
            self.fsid, self.pool, self.image, self.snapshot = \
                map(urllib.parse.unquote, pieces)
        else:
            reason = _('URI must have exactly 1 or 4 components')
            msg = _LI("Invalid URI: %s") % reason

            LOG.info(msg)
            raise exceptions.BadStoreUri(message=reason)
        if any(map(lambda p: p == '', pieces)):
            reason = _('URI cannot contain empty components')
            msg = _LI("Invalid URI: %s") % reason

            LOG.info(msg)
            raise exceptions.BadStoreUri(message=reason)


class ImageIterator(object):
    """
    Reads data from an RBD image, one chunk at a time.
    """

    def __init__(self, pool, name, snapshot, store, chunk_size=None):
        self.pool = pool or store.pool
        self.name = name
        self.snapshot = snapshot
        self.user = store.user
        self.conf_file = store.conf_file
        self.chunk_size = chunk_size or store.READ_CHUNKSIZE
        self.store = store

    def __iter__(self):
        try:
            with self.store.get_connection(conffile=self.conf_file,
                                           rados_id=self.user) as conn:
                with conn.open_ioctx(self.pool) as ioctx:
                    with rbd.Image(ioctx, self.name,
                                   snapshot=self.snapshot,
                                   read_only=True) as image:
                        size = image.size()
                        bytes_left = size
                        while bytes_left > 0:
                            length = min(self.chunk_size, bytes_left)
                            data = image.read(size - bytes_left, length)
                            bytes_left -= len(data)
                            yield data
                        return
        except rbd.ImageNotFound:
            raise exceptions.NotFound(
                _('RBD image %s does not exist') % self.name)


class Store(driver.Store):
    """An implementation of the RBD backend adapter."""

    _CAPABILITIES = capabilities.BitMasks.RW_ACCESS
    OPTIONS = _RBD_OPTS

    EXAMPLE_URL = "rbd://<FSID>/<POOL>/<IMAGE>/<SNAP>"

    def __init__(self, *args, **kargs):
        super(Store, self).__init__(*args, **kargs)
        if self.backend_group:
            self.store_conf = driver.BackendGroupConfiguration(
                self.OPTIONS, self.backend_group, conf=self.conf)
        else:
            self.store_conf = self.conf.glance_store

    def get_schemes(self):
        return ('rbd',)

    def RBDProxy(self):
        if eventletutils.is_monkey_patched('thread'):
            return tpool.Proxy(rbd.RBD())
        else:
            return rbd.RBD()

    @contextlib.contextmanager
    def get_connection(self, conffile, rados_id):
        client = rados.Rados(conffile=conffile, rados_id=rados_id)

        timeout = self.store_conf.rados_connect_timeout

        if timeout >= 0:
            t = str(timeout)
            client.conf_set('rados_osd_op_timeout', t)
            client.conf_set('rados_mon_op_timeout', t)
            client.conf_set('client_mount_timeout', t)

        try:
            client.connect()
        except (rados.Error, rados.ObjectNotFound) as e:
            if self.backend_group and len(self.conf.enabled_backends) > 1:
                reason = _("Error in store configuration: %s") % e
                LOG.debug(reason)
                raise exceptions.BadStoreConfiguration(
                    store_name=self.backend_group, reason=reason)
            else:
                msg = _LE("Error connecting to ceph cluster.")
                LOG.exception(msg)
                raise exceptions.BackendException()
        try:
            yield client
        finally:
            client.shutdown()

    def configure_add(self):
        """
        Configure the Store to use the stored configuration options
        Any store that needs special configuration should implement
        this method. If the store was not able to successfully configure
        itself, it should raise `exceptions.BadStoreConfiguration`
        """
        if rbd is None or rados is None:
            reason = _("The required libraries(rbd and rados) are not "
                       "available")
            LOG.error(reason)
            raise exceptions.BadStoreConfiguration(store_name='rbd',
                                                   reason=reason)

        try:
            chunk = self.store_conf.rbd_store_chunk_size
            pool = self.store_conf.rbd_store_pool
            user = self.store_conf.rbd_store_user
            conf_file = self.store_conf.rbd_store_ceph_conf
            thin_provisioning = self.store_conf.rbd_thin_provisioning

            self.thin_provisioning = thin_provisioning
            self.chunk_size = chunk * units.Mi
            self.READ_CHUNKSIZE = self.chunk_size
            self.WRITE_CHUNKSIZE = self.READ_CHUNKSIZE

            # these must not be unicode since they will be passed to a
            # non-unicode-aware C library
            self.pool = str(pool)
            self.user = str(user)
            self.conf_file = str(conf_file)
        except cfg.ConfigFileValueError as e:
            reason = _("Error in store configuration: %s") % e
            LOG.error(reason)
            raise exceptions.BadStoreConfiguration(store_name='rbd',
                                                   reason=reason)
        if self.backend_group:
            self._set_url_prefix()
        self.size = 0
        self.resize_amount = self.WRITE_CHUNKSIZE

    def _set_url_prefix(self):
        fsid = None
        with self.get_connection(conffile=self.conf_file,
                                 rados_id=self.user) as conn:
            if hasattr(conn, 'get_fsid'):
                fsid = conn.get_fsid()

        if fsid and self.pool:
            # ensure nothing contains / or any other url-unsafe character
            safe_fsid = urllib.parse.quote(fsid, '')
            safe_pool = urllib.parse.quote(self.pool, '')
            self._url_prefix = "rbd://%s/%s/" % (safe_fsid, safe_pool)
        else:
            self._url_prefix = "rbd://"

    @capabilities.check
    def get(self, location, offset=0, chunk_size=None, context=None):
        """
        Takes a `glance_store.location.Location` object that indicates
        where to find the image file, and returns a tuple of generator
        (for reading the image file) and image_size

        :param location: `glance_store.location.Location` object, supplied
                        from glance_store.location.get_location_from_uri()
        :raises: `glance_store.exceptions.NotFound` if image does not exist
        """
        loc = location.store_location
        return (ImageIterator(loc.pool, loc.image, loc.snapshot, self),
                self.get_size(location))

    def get_size(self, location, context=None):
        """
        Takes a `glance_store.location.Location` object that indicates
        where to find the image file, and returns the size

        :param location: `glance_store.location.Location` object, supplied
                        from glance_store.location.get_location_from_uri()
        :raises: `glance_store.exceptions.NotFound` if image does not exist
        """
        loc = location.store_location
        # if there is a pool specific in the location, use it; otherwise
        # we fall back to the default pool specified in the config
        target_pool = loc.pool or self.pool
        with self.get_connection(conffile=self.conf_file,
                                 rados_id=self.user) as conn:
            with conn.open_ioctx(target_pool) as ioctx:
                try:
                    with rbd.Image(ioctx, loc.image,
                                   snapshot=loc.snapshot,
                                   read_only=True) as image:
                        img_info = image.stat()
                        return img_info['size']
                except rbd.ImageNotFound:
                    msg = _('RBD image %s does not exist') % loc.get_uri()
                    LOG.debug(msg)
                    raise exceptions.NotFound(msg)

    def _create_image(self, fsid, conn, ioctx, image_name,
                      size, order, context=None):
        """
        Create an rbd image. If librbd supports it,
        make it a cloneable snapshot, so that copy-on-write
        volumes can be created from it.

        :param image_name: Image's name

        :returns: `glance_store.rbd.StoreLocation` object
        """
        features = conn.conf_get('rbd_default_features')
        if ((features is None) or (int(features) == 0)):
            features = rbd.RBD_FEATURE_LAYERING
        self.RBDProxy().create(ioctx, image_name, size, order,
                               old_format=False,
                               features=int(features))
        return StoreLocation({
            'fsid': fsid,
            'pool': self.pool,
            'image': image_name,
            'snapshot': DEFAULT_SNAPNAME,
        }, self.conf)

    def _delete_image(self, target_pool, image_name,
                      snapshot_name=None, context=None):
        """
        Delete RBD image and snapshot.

        :param image_name: Image's name
        :param snapshot_name: Image snapshot's name

        :raises: NotFound if image does not exist;
                InUseByStore if image is in use or snapshot unprotect failed
        """
        with self.get_connection(conffile=self.conf_file,
                                 rados_id=self.user) as conn:
            with conn.open_ioctx(target_pool) as ioctx:
                try:
                    # First remove snapshot.
                    if snapshot_name is not None:
                        with rbd.Image(ioctx, image_name) as image:
                            try:
                                self._unprotect_snapshot(image, snapshot_name)
                                image.remove_snap(snapshot_name)
                            except rbd.ImageNotFound as exc:
                                msg = (_("Snap Operating Exception "
                                         "%(snap_exc)s "
                                         "Snapshot does not exist.") %
                                       {'snap_exc': exc})
                                LOG.debug(msg)
                            except rbd.ImageBusy as exc:
                                log_msg = (_LW("Snap Operating Exception "
                                               "%(snap_exc)s "
                                               "Snapshot is in use.") %
                                           {'snap_exc': exc})
                                LOG.warning(log_msg)
                                raise exceptions.InUseByStore()

                    # Then delete image.
                    self.RBDProxy().remove(ioctx, image_name)
                except rbd.ImageHasSnapshots:
                    log_msg = (_LW("Unable to remove image %(img_name)s: it "
                                   "has snapshot(s) left; trashing instead") %
                               {'img_name': image_name})
                    LOG.warning(log_msg)
                    with rbd.Image(ioctx, image_name) as image:
                        try:
                            rbd.RBD().trash_move(ioctx, image_name)
                            LOG.debug('Moved %s to trash', image_name)
                        except rbd.ImageBusy:
                            LOG.warning(_('Unable to move in-use image to '
                                          'trash'))
                            raise exceptions.InUseByStore()
                        return
                    raise exceptions.HasSnapshot()
                except rbd.ImageBusy:
                    log_msg = (_LW("Remove image %(img_name)s failed. "
                                   "It is in use.") %
                               {'img_name': image_name})
                    LOG.warning(log_msg)
                    raise exceptions.InUseByStore()
                except rbd.ImageNotFound:
                    msg = _("RBD image %s does not exist") % image_name
                    raise exceptions.NotFound(message=msg)

    def _unprotect_snapshot(self, image, snap_name):
        try:
            image.unprotect_snap(snap_name)
        except rbd.InvalidArgument:
            # NOTE(slaweq): if snapshot was unprotected already, rbd library
            # raises InvalidArgument exception without any "clear" message.
            # Such exception is not dangerous for us so it will be just logged
            LOG.debug("Snapshot %s is unprotected already" % snap_name)

    def _resize_on_write(self, image, image_size, bytes_written, chunk_length):
        """Handle the rbd resize when needed."""
        if image_size != 0 or self.size >= bytes_written + chunk_length:
            return self.size
        # Note(jokke): We double how much we grow the image each time
        # up to 8gigs to avoid resizing for each write on bigger images
        self.resize_amount = min(self.resize_amount * 2, 8 * units.Gi)
        new_size = self.size + self.resize_amount
        LOG.debug("resizing image to %s KiB" % (new_size / units.Ki))
        image.resize(new_size)
        return new_size

    @driver.back_compat_add
    @capabilities.check
    def add(self, image_id, image_file, image_size, hashing_algo, context=None,
            verifier=None):
        """
        Stores an image file with supplied identifier to the backend
        storage system and returns a tuple containing information
        about the stored image.

        :param image_id: The opaque image identifier
        :param image_file: The image data to write, as a file-like object
        :param image_size: The size of the image data to write, in bytes
        :param hashing_algo: A hashlib algorithm identifier (string)
        :param context: A context object
        :param verifier: An object used to verify signatures for images

        :returns: tuple of: (1) URL in backing store, (2) bytes written,
                  (3) checksum, (4) multihash value, and (5) a dictionary
                  with storage system specific information
        :raises: `glance_store.exceptions.Duplicate` if the image already
                 exists
        """
        os_hash_value = utils.get_hasher(hashing_algo, False)
        checksum = utils.get_hasher('md5', False)
        image_name = str(image_id)
        with self.get_connection(conffile=self.conf_file,
                                 rados_id=self.user) as conn:
            fsid = None
            if hasattr(conn, 'get_fsid'):
                fsid = conn.get_fsid()
            with conn.open_ioctx(self.pool) as ioctx:
                order = int(math.log(self.WRITE_CHUNKSIZE, 2))
                LOG.debug('creating image %s with order %d and size %d',
                          image_name, order, image_size)
                if image_size == 0:
                    LOG.warning(_LW("Since image size is zero we will be "
                                    "doing resize-before-write which will be "
                                    "slower than normal"))

                try:
                    loc = self._create_image(fsid, conn, ioctx, image_name,
                                             image_size, order)
                except rbd.ImageExists:
                    msg = _('RBD image %s already exists') % image_id
                    raise exceptions.Duplicate(message=msg)

                try:
                    with rbd.Image(ioctx, image_name) as image:
                        bytes_written = 0
                        offset = 0
                        chunks = utils.chunkreadable(image_file,
                                                     self.WRITE_CHUNKSIZE)
                        for chunk in chunks:
                            # NOTE(jokke): If we don't know image size we need
                            # to resize it on write. The resize amount will
                            # ramp up to 8 gigs.
                            chunk_length = len(chunk)
                            self.size = self._resize_on_write(image,
                                                              image_size,
                                                              bytes_written,
                                                              chunk_length)
                            bytes_written += chunk_length
                            # Check if writing this chunk exceeds the
                            # image_size
                            if image_size != 0 and bytes_written > image_size:
                                raise exceptions.Invalid(
                                    _("Size exceeds: expected %(expected)d "
                                      "bytes, got %(actual)d bytes") %
                                    {'expected': image_size,
                                     'actual': bytes_written})
                            if not (self.thin_provisioning and not any(chunk)):
                                image.write(chunk, offset)
                            offset += chunk_length
                            os_hash_value.update(chunk)
                            checksum.update(chunk)
                            if verifier:
                                verifier.update(chunk)

                        # Lets trim the image in case we overshoot with resize
                        if image_size == 0:
                            image.resize(bytes_written)

                        if loc.snapshot:
                            image.create_snap(loc.snapshot)
                            image.protect_snap(loc.snapshot)
                except rbd.NoSpace:
                    log_msg = (_LE("Failed to store image %(img_name)s "
                                   "insufficient space available") %
                               {'img_name': image_name})
                    LOG.error(log_msg)

                    # Delete image if one was created
                    try:
                        target_pool = loc.pool or self.pool
                        self._delete_image(target_pool, loc.image,
                                           loc.snapshot)
                    except exceptions.NotFound:
                        pass

                    raise exceptions.StorageFull(message=log_msg)
                except rbd.IncompleteWriteError as exc:
                    log_msg = (_LE("Failed to store image %(img_name)s "
                                   "image data exceeded the expected size. "
                                   "Store Exception %(store_exc)s") %
                               {'img_name': image_name,
                                'store_exc': exc})
                    LOG.error(log_msg)

                    # Delete image if one was created
                    try:
                        target_pool = loc.pool or self.pool
                        self._delete_image(target_pool, loc.image,
                                           loc.snapshot)
                    except exceptions.NotFound:
                        pass

                    raise exceptions.Invalid(message=log_msg)
                except Exception as exc:
                    log_msg = (_LE("Failed to store image %(img_name)s "
                                   "Store Exception %(store_exc)s") %
                               {'img_name': image_name,
                                'store_exc': exc})
                    LOG.error(log_msg)

                    # Delete image if one was created
                    try:
                        target_pool = loc.pool or self.pool
                        self._delete_image(target_pool, loc.image,
                                           loc.snapshot)
                    except exceptions.NotFound:
                        pass

                    raise exc

        # If actual bytes written are less than image_size
        # (when image_size != 0), update accordingly
        if image_size != 0 and bytes_written != image_size:
            # Delete image if one was created
            try:
                self._delete_image(loc.pool or self.pool, loc.image,
                                   loc.snapshot)
            except exceptions.NotFound:
                pass

            raise exceptions.Invalid(_(
                "Size mismatch: expected %(expected)d "
                "bytes, got %(actual)d bytes") % {
                'expected': image_size, 'actual': bytes_written})

        # Add store backend information to location metadata
        metadata = {}
        if self.backend_group:
            metadata['store'] = self.backend_group

        return (loc.get_uri(),
                bytes_written,
                checksum.hexdigest(),
                os_hash_value.hexdigest(),
                metadata)

    @capabilities.check
    def delete(self, location, context=None):
        """
        Takes a `glance_store.location.Location` object that indicates
        where to find the image file to delete.

        :param location: `glance_store.location.Location` object, supplied
                  from glance_store.location.get_location_from_uri()

        :raises: NotFound if image does not exist;
                InUseByStore if image is in use or snapshot unprotect failed
        """
        loc = location.store_location
        target_pool = loc.pool or self.pool
        self._delete_image(target_pool, loc.image, loc.snapshot)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests the filesystem backend store"""

import errno
import io
import json
import os
import threading
import time
import uuid

from unittest import mock

import futurist
from oslo_utils import units

from glance_store._drivers import filesystem
from glance_store import exceptions
from glance_store import location
from glance_store.tests import base
from glance_store.tests.unit import test_filesystem_store_base as file_base
from glance_store.tests.unit import test_store_capabilities


class TestStore(base.StoreBaseTest,
                file_base.TestFilerStoreBase,
                test_store_capabilities.TestStoreCapabilitiesChecking):

    def setUp(self):
        super(TestStore, self).setUp()
        # Set default values for multistore and backend
        self.multistore = False
        self.backend = 'glance_store'

        self.store = filesystem.Store(self.conf)
        self.config(filesystem_store_datadir=self.test_dir,
                    filesystem_store_chunk_size=10,
                    stores=['glance.store.filesystem.Store'],
                    group="glance_store")
        self.store.configure()
        self.register_store_schemes(self.store, 'file')
        self.hash_algo = 'sha256'

    def _create_metadata_json_file(self, metadata, group=None):
        expected_image_id = str(uuid.uuid4())
        jsonfilename = os.path.join(self.test_dir,
                                    "storage_metadata.%s" % expected_image_id)
        self.config(filesystem_store_metadata_file=jsonfilename,
                    group=group)
        with open(jsonfilename, 'w') as fptr:
            json.dump(metadata, fptr)

    def _store_image(self, in_metadata):
        expected_image_id = str(uuid.uuid4())
        expected_file_size = 10
        expected_file_contents = b"*" * expected_file_size
        image_file = io.BytesIO(expected_file_contents)
        self.store.FILESYSTEM_STORE_METADATA = in_metadata
        return self.store.add(expected_image_id, image_file,
                              expected_file_size, self.hash_algo)

    def test_get(self):
        self._test_get()

    def test_get_random_access(self):
        self._test_get_random_access()

    def test_get_non_existing(self):
        """
        Test that trying to retrieve a file that doesn't exist
        raises an error
        """
        self._test_get_non_existing()

    def test_add(self):
        self._test_add()

    def test_add_image_exceeding_max_size_raises_exception(self):
        self._test_add_image_exceeding_max_size_raises_exception()

    def test_write_less_than_declared_raises_exception(self):
        self._test_write_less_than_declared_raises_exception()

    def test_thin_provisioning_is_disabled_by_default(self):
        self.assertEqual(self.store.thin_provisioning, False)

    def test_add_with_thick_provisioning(self):
        self._do_test_add(enable_thin_provisoning=False)

    def test_add_with_thin_provisioning(self):
        self._do_test_add(enable_thin_provisoning=True)

    def test_add_thick_provisioning_with_holes_in_file(self):
        """
        Tests that a file which contains null bytes chunks is fully
        written with a thick provisioning configuration.
        """
        chunk_size = units.Ki  # 1K
        content = b"*" * chunk_size + b"\x00" * chunk_size + b"*" * chunk_size
        self._do_test_thin_provisioning(content, 3 * chunk_size, 0, 3, False)

    def test_add_thin_provisioning_with_holes_in_file(self):
        """
        Tests that a file which contains null bytes chunks is sparsified
        with a thin provisioning configuration.
        """
        chunk_size = units.Ki  # 1K
        content = b"*" * chunk_size + b"\x00" * chunk_size + b"*" * chunk_size
        self._do_test_thin_provisioning(content, 3 * chunk_size, 1, 2, True)

    def test_add_thick_provisioning_without_holes_in_file(self):
        """
        Tests that a file which not contain null bytes chunks is fully
        written with a thick provisioning configuration.
        """
        chunk_size = units.Ki  # 1K
        content = b"*" * 3 * chunk_size
        self._do_test_thin_provisioning(content, 3 * chunk_size, 0, 3, False)

    def test_add_thin_provisioning_without_holes_in_file(self):
        """
        Tests that a file which not contain null bytes chunks is fully
        written with a thin provisioning configuration.
        """
        chunk_size = units.Ki  # 1K
        content = b"*" * 3 * chunk_size
        self._do_test_thin_provisioning(content, 3 * chunk_size, 0, 3, True)

    def test_add_thick_provisioning_with_partial_holes_in_file(self):
        """
        Tests that a file which contains null bytes not aligned with
        chunk size is fully written with a thick provisioning configuration.
        """
        chunk_size = units.Ki  # 1K
        my_chunk = int(chunk_size * 1.5)
        content = b"*" * my_chunk + b"\x00" * my_chunk + b"*" * my_chunk
        self._do_test_thin_provisioning(content, 3 * my_chunk, 0, 5, False)

    def test_add_thin_provisioning_with_partial_holes_in_file(self):
        """
        Tests that a file which contains null bytes not aligned with
        chunk size is sparsified with a thin provisioning configuration.
        """
        chunk_size = units.Ki  # 1K
        my_chunk = int(chunk_size * 1.5)
        content = b"*" * my_chunk + b"\x00" * my_chunk + b"*" * my_chunk
        self._do_test_thin_provisioning(content, 3 * my_chunk, 1, 4, True)

    def test_add_with_verifier(self):
        self._test_add_with_verifier()

    def test_add_check_metadata_with_invalid_mountpoint_location(self):
        self._test_add_check_metadata_with_invalid_mountpoint_location()

    def test_add_check_metadata_list_with_invalid_mountpoint_locations(self):
        self._test_add_check_metadata_list_with_invalid_mountpoint_locations()

    def test_add_check_metadata_list_with_valid_mountpoint_locations(self):
        self._test_add_check_metadata_list_with_valid_mountpoint_locations()

    def test_add_check_metadata_bad_nosuch_file(self):
        self._test_add_check_metadata_bad_nosuch_file()

    def test_add_already_existing(self):
        self._test_add_already_existing()

    def test_add_storage_full(self):
        """
        Tests that adding an image without enough space on disk
        raises an appropriate exception
        """
        self._do_test_add_write_failure(errno.ENOSPC, exceptions.StorageFull)

    def test_add_file_too_big(self):
        """
        Tests that adding an excessively large image file
        raises an appropriate exception
        """
        self._do_test_add_write_failure(errno.EFBIG, exceptions.StorageFull)

    def test_add_storage_write_denied(self):
        """
        Tests that adding an image with insufficient filestore permissions
        raises an appropriate exception
        """
        self._do_test_add_write_failure(errno.EACCES,
                                        exceptions.StorageWriteDenied)

    def test_add_other_failure(self):
        """
        Tests that a non-space-related IOError does not raise a
        StorageFull exceptions.
        """
        self._do_test_add_write_failure(errno.ENOTDIR, IOError)

    def test_add_cleanup_on_read_failure(self):
        self._test_add_cleanup_on_read_failure()

    def test_delete(self):
        self._test_delete()

    def test_delete_non_existing(self):
        self._test_delete_non_existing()

    def test_delete_forbidden(self):
        self._test_delete_forbidden()

    def test_configure_add_with_multi_datadirs(self):
        self._test_configure_add_with_multi_datadirs()

    def test_configure_add_with_metadata_file_success(self):
        self._test_configure_add_with_metadata_file_success()

    def test_configure_add_check_metadata_list_of_dicts_success(self):
        self._test_configure_add_check_metadata_list_of_dicts_success()

    def test_configure_add_check_metadata_success_list_val_for_some_key(self):
        self._test_configure_add_check_metadata_success_list_val_for_some_key()

    def test_configure_add_check_metadata_bad_data(self):
        self._test_configure_add_check_metadata_bad_data()

    def test_configure_add_check_metadata_with_no_id_or_mountpoint(self):
        self._test_configure_add_check_metadata_with_no_id_or_mountpoint()

    def test_configure_add_check_metadata_id_or_mountpoint_is_not_string(self):
        self._test_cfg_add_check_metadata_id_or_mountpoint_is_not_string()

    def test_configure_add_check_metadata_list_with_no_id_or_mountpoint(self):
        self._test_cfg_add_check_metadata_list_with_no_id_or_mountpoint()

    def test_add_check_metadata_list_id_or_mountpoint_is_not_string(self):
        self._test_add_check_metadata_list_id_or_mountpoint_is_not_string()

    def test_configure_add_same_dir_multiple_times(self):
        self._test_configure_add_same_dir_multiple_times()

    def test_configure_add_same_dir_multiple_times_same_priority(self):
        self._test_configure_add_same_dir_multiple_times_same_priority()

    def test_add_with_multiple_dirs(self):
        self._test_add_with_multiple_dirs()

    def test_add_with_multiple_dirs_storage_full(self):
        self._test_add_with_multiple_dirs_storage_full()

    def test_configure_add_with_file_perm(self):
        self._test_configure_add_with_file_perm()

    def test_configure_add_with_inaccessible_file_perm(self):
        self._test_configure_add_with_inaccessible_file_perm()

    def test_add_with_file_perm_for_group_other_users_access(self):
        self._test_add_with_file_perm_for_group_other_users_access()

    def test_add_with_file_perm_for_owner_users_access(self):
        self._test_add_with_file_perm_for_owner_users_access()

    def test_configure_add_chunk_size(self):
        self._test_configure_add_chunk_size()

    def test_timeout_executor_disabled_by_default(self):
        self.store.configure()
        self.assertEqual(self.store.timeout_executor.timeout, 0)
        self.assertIsInstance(self.store.timeout_executor.executor,
                              futurist.SynchronousExecutor)

    def test_timeout_executor_enabled_with_timeout(self):
        self.config(filesystem_store_timeout=30,
                    filesystem_store_thread_pool_size=5,
                    filesystem_store_threadpool_threshold=80,
                    group="glance_store")
        self.store.configure()
        self.assertEqual(self.store.timeout_executor.timeout, 30)
        self.assertEqual(self.store.timeout_executor.pool_size, 5)
        self.assertEqual(self.store.timeout_executor.threshold, 80)
        self.assertIsNotNone(self.store.timeout_executor.executor)

    def test_delete_with_timeout_success(self):
        image_id = str(uuid.uuid4())
        file_contents = b"test content"
        image_file = io.BytesIO(file_contents)
        uri, size, checksum, multihash, _ = self.store.add(
            image_id, image_file, len(file_contents), self.hash_algo)
        self.config(filesystem_store_timeout=30, group="glance_store")
        self.store.configure()
        loc = location.get_location_from_uri(uri, conf=self.conf)
        self.store.delete(loc)
        loc2 = location.get_location_from_uri(uri, conf=self.conf)
        self.assertRaises(exceptions.NotFound, self.store.get, loc2)

    def test_delete_with_timeout_failure(self):
        image_id = str(uuid.uuid4())
        file_contents = b"test content"
        image_file = io.BytesIO(file_contents)
        uri, size, checksum, multihash, _ = self.store.add(
            image_id, image_file, len(file_contents), self.hash_algo)
        self.config(filesystem_store_timeout=1, group="glance_store")
        self.store.configure()
        loc = location.get_location_from_uri(uri, conf=self.conf)
        with mock.patch('os.unlink') as mock_unlink:

            def slow_unlink(path):
                time.sleep(2)

            mock_unlink.side_effect = slow_unlink
            self.assertRaises(exceptions.TimeoutError, self.store.delete, loc)

    def test_get_size_with_timeout_success(self):
        image_id = str(uuid.uuid4())
        file_contents = b"test content"
        image_file = io.BytesIO(file_contents)
        uri, size, checksum, multihash, _ = self.store.add(
            image_id, image_file, len(file_contents), self.hash_algo)
        self.config(filesystem_store_timeout=30, group="glance_store")
        self.store.configure()
        loc = location.get_location_from_uri(uri, conf=self.conf)
        result_size = self.store.get_size(loc)
        self.assertEqual(result_size, len(file_contents))

    def test_get_size_with_timeout_failure(self):
        image_id = str(uuid.uuid4())
        file_contents = b"test content"
        image_file = io.BytesIO(file_contents)
        uri, size, checksum, multihash, _ = self.store.add(
            image_id, image_file, len(file_contents), self.hash_algo)
        self.config(filesystem_store_timeout=1, group="glance_store")
        self.store.configure()
        loc = location.get_location_from_uri(uri, conf=self.conf)
        with mock.patch('os.path.getsize') as mock_getsize:
            def slow_getsize(path):
                time.sleep(2)
                return len(file_contents)
            mock_getsize.side_effect = slow_getsize
            self.assertRaises(exceptions.TimeoutError,
                              self.store.get_size, loc)

    def test_get_capacity_info_with_timeout_success(self):
        self.config(filesystem_store_timeout=30, group="glance_store")
        self.store.configure()
        capacity = self.store._get_capacity_info(self.test_dir)
        self.assertGreaterEqual(capacity, 0)

    def test_get_capacity_info_with_timeout_failure(self):
        self.config(filesystem_store_timeout=1, group="glance_store")
        self.store.configure()
        with mock.patch('os.statvfs') as mock_statvfs:

            def slow_statvfs(path):
                time.sleep(2)

                class FakeStatvfs:
                    f_bavail = 1000
                    f_bsize = 4096

                return FakeStatvfs()

            mock_statvfs.side_effect = slow_statvfs
            self.assertRaises(exceptions.TimeoutError,
                              self.store._get_capacity_info,
                              self.test_dir)

    def test_timeout_executor_no_overhead_when_disabled(self):
        image_id = str(uuid.uuid4())
        file_contents = b"test content"
        image_file = io.BytesIO(file_contents)
        uri, size, checksum, multihash, _ = self.store.add(
            image_id, image_file, len(file_contents), self.hash_algo)
        self.config(filesystem_store_timeout=0, group="glance_store")
        self.store.configure()
        self.assertIsInstance(self.store.timeout_executor.executor,
                              futurist.SynchronousExecutor)
        loc = location.get_location_from_uri(uri, conf=self.conf)
        result_size = self.store.get_size(loc)
        self.assertEqual(result_size, len(file_contents))


class TestTimeoutExecutor(base.StoreBaseTest):
    """Test TimeoutExecutor utility class"""

    def test_thread_pool_threshold_warning(self):
        """Test that warning is logged when thread pool usage exceeds
        threshold
        """
        pool_size = 3
        threshold = 66
        timeout = 10

        executor = filesystem.TimeoutExecutor(timeout, pool_size, threshold)
        event = threading.Event()

        def blocking_func():
            event.wait()

        with mock.patch.object(filesystem.LOG, 'warning') as mock_warning:
            def worker():
                executor.execute(blocking_func)

            threads = []
            for i in range(2):
                t = threading.Thread(target=worker)
                t.start()
                threads.append(t)

            time.sleep(0.2)
            executor.execute(lambda: 1)

            mock_warning.assert_called_once()
            call_args = mock_warning.call_args[0][0]
            self.assertIn('Thread pool usage', call_args)
            self.assertIn('threshold', call_args)
            self.assertIn('%d%%' % threshold, call_args)
            self.assertIn('Pool may start blocking', call_args)

            event.set()
            for t in threads:
                t.join(timeout=5)

        executor.shutdown(wait=True)

    def test_timeout_behavior(self):
        """Test that operations timeout cleanly when timeout is exceeded"""
        pool_size = 1
        threshold = 80
        timeout = 0.1

        executor = filesystem.TimeoutExecutor(timeout, pool_size, threshold)
        event = threading.Event()

        def blocking_func():
            event.wait()

        self.assertRaises(exceptions.TimeoutError,
                          executor.execute, blocking_func)

        event.set()

        executor.shutdown(wait=True)



# Copyright 2025 RedHat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import io
from unittest import mock
import uuid

from oslo_utils import units

from glance_store._drivers import rbd as rbd_store
from glance_store import exceptions
from glance_store import location
from glance_store.tests import utils as test_utils


class TestException(Exception):
    pass


class MockRados(object):

    class Error(Exception):
        pass

    class ObjectNotFound(Exception):
        pass

    class ioctx(object):
        def __init__(self, *args, **kwargs):
            pass

        def __enter__(self, *args, **kwargs):
            return self

        def __exit__(self, *args, **kwargs):
            return False

        def close(self, *args, **kwargs):
            pass

    class Rados(object):

        def __init__(self, *args, **kwargs):
            pass

        def __enter__(self, *args, **kwargs):
            return self

        def __exit__(self, *args, **kwargs):
            return False

        def connect(self, *args, **kwargs):
            pass

        def open_ioctx(self, *args, **kwargs):
            return MockRados.ioctx()

        def shutdown(self, *args, **kwargs):
            pass

        def conf_get(self, *args, **kwargs):
            pass


class MockRBD(object):

    class ImageExists(Exception):
        pass

    class ImageHasSnapshots(Exception):
        pass

    class IncompleteWriteError(Exception):
        pass

    class ImageBusy(Exception):
        pass

    class ImageNotFound(Exception):
        pass

    class InvalidArgument(Exception):
        pass

    class NoSpace(Exception):
        pass

    class Image(object):

        def __init__(self, *args, **kwargs):
            pass

        def __enter__(self, *args, **kwargs):
            return self

        def __exit__(self, *args, **kwargs):
            pass

        def create_snap(self, *args, **kwargs):
            pass

        def remove_snap(self, *args, **kwargs):
            pass

        def set_snap(self, *args, **kwargs):
            pass

        def list_children(self, *args, **kwargs):
            pass

        def protect_snap(self, *args, **kwargs):
            pass

        def unprotect_snap(self, *args, **kwargs):
            pass

        def read(self, *args, **kwargs):
            raise NotImplementedError()

        def write(self, *args, **kwargs):
            raise NotImplementedError()

        def resize(self, *args, **kwargs):
            pass

        def discard(self, offset, length):
            raise NotImplementedError()

        def close(self):
            pass

        def list_snaps(self):
            raise NotImplementedError()

        def parent_info(self):
            raise NotImplementedError()

        def size(self):
            raise NotImplementedError()

    class RBD(object):

        def __init__(self, *args, **kwargs):
            pass

        def __enter__(self, *args, **kwargs):
            return self

        def __exit__(self, *args, **kwargs):
            return False

        def create(self, *args, **kwargs):
            pass

        def remove(self, *args, **kwargs):
            pass

        def list(self, *args, **kwargs):
            raise NotImplementedError()

        def clone(self, *args, **kwargs):
            raise NotImplementedError()

        def trash_move(self, *args, **kwargs):
            pass

    RBD_FEATURE_LAYERING = 1


class TestRBDStoreBase(object):

    def _test_add(self):
        """Test that we can add an image via the RBD backend."""
        expected_image_id = str(uuid.uuid4())
        expected_file_size = 5 * units.Ki  # 5K
        expected_file_contents = b"*" * expected_file_size
        expected_checksum = hashlib.md5(expected_file_contents,
                                        usedforsecurity=False).hexdigest()
        expected_multihash = hashlib.sha256(expected_file_contents).hexdigest()
        image_file = io.BytesIO(expected_file_contents)

        with mock.patch.object(rbd_store.rbd.Image, 'write'):
            if self.multistore:
                loc, size, checksum, metadata = self.store.add(
                    expected_image_id, image_file, expected_file_size)
                self.assertEqual(self.backend, metadata['store'])
            else:
                loc, size, checksum, multihash, _ = self.store.add(
                    expected_image_id, image_file, expected_file_size,
                    self.hash_algo)
                self.assertEqual(expected_multihash, multihash)

        self.assertEqual(expected_file_size, size)
        self.assertEqual(expected_checksum, checksum)

    def _test_add_image_exceeding_max_size_raises_exception(self):
        """Test that adding an image exceeding max size raises exception."""
        expected_image_id = str(uuid.uuid4())
        total_bytes = units.Ki * 5
        # actual data size
        image_size = total_bytes
        # Create data larger than image_size to simulate mismatch at end
        data = b'a' * (total_bytes + 1024)
        image_file = io.BytesIO(data)

        with mock.patch.object(
                rbd_store.Store, '_delete_image') as mock_delete:
            with mock.patch.object(rbd_store.rbd.Image, 'write'):
                if self.multistore:
                    self.assertRaisesRegex(
                        exceptions.Invalid, "Size exceeds: expected",
                        self.store.add, expected_image_id, image_file,
                        image_size)
                else:
                    self.assertRaisesRegex(
                        exceptions.Invalid, "Size exceeds: expected",
                        self.store.add, expected_image_id, image_file,
                        image_size, self.hash_algo)

                # Confirm that image deletion was called due to size mismatch
                mock_delete.assert_called()
                # The position should be equal to total input size
                self.assertEqual(image_file.tell(), len(data))

    def _test_write_less_than_declared_raises_exception(self):
        """Test that writing less than declared raises exception."""
        expected_image_id = str(uuid.uuid4())
        total_bytes = units.Ki * 5
        # actual data size
        image_size = total_bytes
        # Create data smaller than image_size to simulate mismatch at end
        data = b'a' * (total_bytes - 100)
        image_file = io.BytesIO(data)

        with mock.patch.object(
                rbd_store.Store, '_delete_image') as mock_delete:
            with mock.patch.object(rbd_store.rbd.Image, 'write'):
                if self.multistore:
                    self.assertRaisesRegex(
                        exceptions.Invalid, "Size mismatch: expected",
                        self.store.add, expected_image_id, image_file,
                        image_size)
                else:
                    self.assertRaisesRegex(
                        exceptions.Invalid, "Size mismatch: expected",
                        self.store.add, expected_image_id, image_file,
                        image_size, self.hash_algo)

                # Confirm that image deletion was called due to size mismatch
                mock_delete.assert_called()
                # The position should be less than total input size
                self.assertEqual(image_file.tell(), len(data))

    def _test_add_with_verifier(self):
        """Test that 'verifier.update' is called when verifier is provided."""
        verifier = mock.MagicMock(name='mock_verifier')
        image_id = str(uuid.uuid4())
        file_size = units.Ki  # 1K
        file_contents = b"*" * file_size
        image_file = io.BytesIO(file_contents)

        with mock.patch.object(rbd_store.rbd.Image, 'write'):
            if self.multistore:
                self.store.add(image_id, image_file, file_size,
                               verifier=verifier)
            else:
                self.store.add(image_id, image_file, file_size,
                               self.hash_algo, verifier=verifier)

        verifier.update.assert_called_with(file_contents)

    def _test_add_duplicate_image(self):
        """Test that adding a duplicate image raises exception."""
        def _fake_create_image(*args, **kwargs):
            self.called_commands_actual.append('create')
            raise MockRBD.ImageExists()

        with mock.patch.object(self.store, '_create_image') as create_image:
            create_image.side_effect = _fake_create_image

            if self.multistore:
                self.assertRaises(exceptions.Duplicate, self.store.add,
                                  'fake_image_id', self.data_iter,
                                  self.data_len)
            else:
                self.assertRaises(exceptions.Duplicate, self.store.add,
                                  'fake_image_id', self.data_iter,
                                  self.data_len, self.hash_algo)
            self.called_commands_expected = ['create']

    def _test_delete(self):
        """Test that we can delete an existing image in the RBD store."""
        def _fake_remove(*args, **kwargs):
            self.called_commands_actual.append('remove')

        with mock.patch.object(MockRBD.RBD, 'remove') as remove_image:
            remove_image.side_effect = _fake_remove

            if self.multistore:
                loc = location.get_location_from_uri_and_backend(
                    self.location.get_uri(), 'ceph1', conf=self.conf)
            else:
                loc = location.Location('test_rbd_store',
                                        rbd_store.StoreLocation,
                                        self.conf,
                                        uri=self.location.get_uri())

            self.store.delete(loc)
            self.called_commands_expected = ['remove']

    def _test_delete_non_existing(self):
        """Test that deleting a non-existing image raises exception."""
        def _fake_remove(*args, **kwargs):
            self.called_commands_actual.append('remove')
            raise MockRBD.ImageNotFound()

        with mock.patch.object(MockRBD.RBD, 'remove') as remove:
            remove.side_effect = _fake_remove
            self.assertRaises(exceptions.NotFound, self.store._delete_image,
                              'fake_pool', self.location.image)

            self.called_commands_expected = ['remove']

    def _test_delete_image_with_snap(self):
        """Test deleting an image with snapshot."""
        @mock.patch.object(MockRBD.RBD, 'remove')
        @mock.patch.object(MockRBD.Image, 'remove_snap')
        @mock.patch.object(MockRBD.Image, 'unprotect_snap')
        def _test_delete_with_snap(unprotect, remove_snap, remove):
            def _fake_unprotect_snap(*args, **kwargs):
                self.called_commands_actual.append('unprotect_snap')

            def _fake_remove_snap(*args, **kwargs):
                self.called_commands_actual.append('remove_snap')

            def _fake_remove(*args, **kwargs):
                self.called_commands_actual.append('remove')

            remove.side_effect = _fake_remove
            unprotect.side_effect = _fake_unprotect_snap
            remove_snap.side_effect = _fake_remove_snap
            self.store._delete_image('fake_pool', self.location.image,
                                     snapshot_name='snap')

            self.called_commands_expected = ['unprotect_snap', 'remove_snap',
                                             'remove']

        _test_delete_with_snap()

    def _test_delete_image_with_unprotected_snap(self):
        """Test deleting an image with unprotected snapshot."""
        @mock.patch.object(MockRBD.RBD, 'remove')
        @mock.patch.object(MockRBD.Image, 'remove_snap')
        @mock.patch.object(MockRBD.Image, 'unprotect_snap')
        def _test_delete_with_unprotected_snap(unprotect, remove_snap, remove):
            def _fake_unprotect_snap(*args, **kwargs):
                self.called_commands_actual.append('unprotect_snap')
                raise MockRBD.InvalidArgument()

            def _fake_remove_snap(*args, **kwargs):
                self.called_commands_actual.append('remove_snap')

            def _fake_remove(*args, **kwargs):
                self.called_commands_actual.append('remove')

            remove.side_effect = _fake_remove
            unprotect.side_effect = _fake_unprotect_snap
            remove_snap.side_effect = _fake_remove_snap
            self.store._delete_image('fake_pool', self.location.image,
                                     snapshot_name='snap')

            self.called_commands_expected = ['unprotect_snap', 'remove_snap',
                                             'remove']

        _test_delete_with_unprotected_snap()

    def _test_delete_image_with_snap_with_error(self):
        """Test deleting an image with snapshot that raises error."""
        @mock.patch.object(MockRBD.RBD, 'remove')
        @mock.patch.object(MockRBD.Image, 'remove_snap')
        @mock.patch.object(MockRBD.Image, 'unprotect_snap')
        def _test_delete_with_snap_error(unprotect, remove_snap, remove):
            def _fake_unprotect_snap(*args, **kwargs):
                self.called_commands_actual.append('unprotect_snap')
                raise TestException()

            def _fake_remove_snap(*args, **kwargs):
                self.called_commands_actual.append('remove_snap')

            def _fake_remove(*args, **kwargs):
                self.called_commands_actual.append('remove')

            remove.side_effect = _fake_remove
            unprotect.side_effect = _fake_unprotect_snap
            remove_snap.side_effect = _fake_remove_snap
            self.assertRaises(TestException, self.store._delete_image,
                              'fake_pool', self.location.image,
                              snapshot_name='snap')

            self.called_commands_expected = ['unprotect_snap']

        _test_delete_with_snap_error()

    def _test_delete_image_with_snap_exc_image_busy(self):
        """Test deleting an image with snapshot that is busy."""
        def _fake_unprotect_snap(*args, **kwargs):
            self.called_commands_actual.append('unprotect_snap')
            raise MockRBD.ImageBusy()

        with mock.patch.object(MockRBD.Image, 'unprotect_snap') as mocked:
            mocked.side_effect = _fake_unprotect_snap

            self.assertRaises(exceptions.InUseByStore,
                              self.store._delete_image,
                              'fake_pool', self.location.image,
                              snapshot_name='snap')

            self.called_commands_expected = ['unprotect_snap']

    def _test_delete_image_snap_has_external_references(self):
        """Test deleting an image with snapshot that has references."""
        with mock.patch.object(MockRBD.Image, 'list_children') as mocked:
            mocked.return_value = True

            self.store._delete_image('fake_pool',
                                     self.location.image,
                                     snapshot_name='snap')

    def _test_delete_image_with_snap_exc_image_has_snap(self):
        """Test deleting an image with snapshot that has snapshots."""
        def _fake_remove(*args, **kwargs):
            self.called_commands_actual.append('remove')
            raise MockRBD.ImageHasSnapshots()

        with mock.patch.object(MockRBD.RBD, 'remove') as remove:
            remove.side_effect = _fake_remove
            self.store._delete_image('fake_pool',
                                     self.location.image)

            self.called_commands_expected = ['remove']

    def _test_get_partial_image(self):
        """Test that getting partial image raises exception."""
        if self.multistore:
            loc = location.get_location_from_uri_and_backend(
                self.location.get_uri(), 'ceph1', conf=self.conf)
        else:
            loc = location.Location('test_rbd_store', rbd_store.StoreLocation,
                                    self.conf, store_specs=self.store_specs)
        self.assertRaises(exceptions.StoreRandomGetNotSupported,
                          self.store.get, loc, chunk_size=1)

    def _test_rados_connect_error(self):
        """Test that rados connect error raises exception."""
        @mock.patch.object(MockRados.Rados, 'connect',
                           side_effect=MockRados.Error)
        def _test_connect_error(_):
            rbd_store.rados.Error = MockRados.Error
            rbd_store.rados.ObjectNotFound = MockRados.ObjectNotFound

            def test():
                with self.store.get_connection('conffile', 'rados_id'):
                    pass
            if self.multistore:
                self.assertRaises(exceptions.BadStoreConfiguration, test)
            else:
                self.assertRaises(exceptions.BackendException, test)

        _test_connect_error()

    def _test_create_image_conf_features(self):
        """Test creating image with configuration features."""
        # Tests that we use non-0 features from ceph.conf and cast to int.
        fsid = 'fake'
        features = '3'
        conf_get_mock = mock.Mock(return_value=features)
        conn = mock.Mock(conf_get=conf_get_mock)
        ioctxt = mock.sentinel.ioctxt
        name = '1'
        size = 1024
        order = 3
        with mock.patch.object(rbd_store.rbd.RBD, 'create') as create_mock:
            location = self.store._create_image(
                fsid, conn, ioctxt, name, size, order)
            self.assertEqual(fsid, location.specs['fsid'])
            self.assertEqual(rbd_store.DEFAULT_POOL, location.specs['pool'])
            self.assertEqual(name, location.specs['image'])
            self.assertEqual(rbd_store.DEFAULT_SNAPNAME,
                             location.specs['snapshot'])

        create_mock.assert_called_once_with(ioctxt, name, size, order,
                                            old_format=False, features=3)

    def _test_add_with_thick_provisioning(self):
        """Test adding image with thick provisioning."""
        chunk_size = units.Mi
        content = b"*" * chunk_size + b"\x00" * chunk_size + b"*" * chunk_size
        self._do_test_thin_provisioning(content, 3 * chunk_size, 3, False)

    def _test_add_with_thin_provisioning(self):
        """Test adding image with thin provisioning."""
        chunk_size = units.Mi
        content = b"*" * chunk_size + b"\x00" * chunk_size + b"*" * chunk_size
        self._do_test_thin_provisioning(content, 3 * chunk_size, 2, True)

    def _test_add_thick_provisioning_without_holes(self):
        """Test adding image with thick provisioning without holes."""
        chunk_size = units.Mi
        content = b"*" * 3 * chunk_size
        self._do_test_thin_provisioning(content, 3 * chunk_size, 3, False)

    def _test_add_thin_provisioning_without_holes(self):
        """Test adding image with thin provisioning without holes."""
        chunk_size = units.Mi
        content = b"*" * 3 * chunk_size
        self._do_test_thin_provisioning(content, 3 * chunk_size, 3, True)

    def _test_add_thick_provisioning_with_partial_holes(self):
        """Test adding image with thick provisioning with partial holes."""
        chunk_size = units.Mi
        my_chunk = int(chunk_size * 1.5)
        content = b"*" * my_chunk + b"\x00" * my_chunk + b"*" * my_chunk
        self._do_test_thin_provisioning(content, 3 * my_chunk, 5, False)

    def _test_add_thin_provisioning_with_partial_holes(self):
        """Test adding image with thin provisioning with partial holes."""
        chunk_size = units.Mi
        my_chunk = int(chunk_size * 1.5)
        content = b"*" * my_chunk + b"\x00" * my_chunk + b"*" * my_chunk
        self._do_test_thin_provisioning(content, 3 * my_chunk, 4, True)

    def _do_test_thin_provisioning(self, content, size, write, thin):
        """Helper method to test thin provisioning."""
        self.config(rbd_store_chunk_size=1,
                    rbd_thin_provisioning=thin,
                    group=self.backend)
        self.store.configure()

        image_id = 'fake_image_id'
        image_file = io.BytesIO(content)
        expected_checksum = hashlib.md5(content,
                                        usedforsecurity=False).hexdigest()
        expected_multihash = hashlib.sha256(content).hexdigest()

        with mock.patch.object(rbd_store.rbd.Image, 'write') as mock_write:
            if self.multistore:
                loc, size, checksum, metadata = self.store.add(
                    image_id, image_file, size)
                self.assertEqual(self.backend, metadata['store'])
            else:
                loc, size, checksum, multihash, _ = self.store.add(
                    image_id, image_file, size, self.hash_algo)
                self.assertEqual(expected_multihash, multihash)
            self.assertEqual(mock_write.call_count, write)

        self.assertEqual(expected_checksum, checksum)

    def _test_add_w_image_size_zero(self):
        """Test that correct size is returned even though 0 was provided."""
        self.store.chunk_size = units.Ki
        with mock.patch.object(rbd_store.rbd.Image, 'resize') as resize:
            with mock.patch.object(rbd_store.rbd.Image, 'write') as write:
                if self.multistore:
                    ret = self.store.add('fake_image_id', self.data_iter, 0)
                    self.assertEqual(self.backend, ret[3]['store'])
                else:
                    ret = self.store.add('fake_image_id', self.data_iter, 0,
                                         self.hash_algo)

                self.assertTrue(resize.called)
                self.assertTrue(write.called)
                self.assertEqual(ret[1], self.data_len)

    def _test_add_w_rbd_image_exception(self):
        """Test adding image with RBD image exception."""
        @mock.patch.object(MockRBD.Image, '__enter__')
        @mock.patch.object(rbd_store.Store, '_create_image')
        @mock.patch.object(rbd_store.Store, '_delete_image')
        def _test_add_with_exception(delete, create, enter):
            def _fake_create_image(*args, **kwargs):
                self.called_commands_actual.append('create')
                return self.location

            def _fake_delete_image(
                    target_pool, image_name, snapshot_name=None):
                self.assertEqual(self.location.pool, target_pool)
                self.assertEqual(self.location.image, image_name)
                self.assertEqual(self.location.snapshot, snapshot_name)
                self.called_commands_actual.append('delete')

            def _fake_enter(*args, **kwargs):
                raise exceptions.NotFound(image="fake_image_id")

            create.side_effect = _fake_create_image
            delete.side_effect = _fake_delete_image
            enter.side_effect = _fake_enter

            if self.multistore:
                self.assertRaises(exceptions.NotFound,
                                  self.store.add,
                                  'fake_image_id', self.data_iter,
                                  self.data_len)
            else:
                self.assertRaises(exceptions.NotFound,
                                  self.store.add,
                                  'fake_image_id', self.data_iter,
                                  self.data_len, self.hash_algo)

            self.called_commands_expected = ['create', 'delete']

        _test_add_with_exception()

    def _test_add_w_rbd_no_space_exception(self):
        """Test adding image with RBD no space exception."""
        @mock.patch.object(MockRBD.Image, 'resize')
        @mock.patch.object(rbd_store.Store, '_create_image')
        @mock.patch.object(rbd_store.Store, '_delete_image')
        def _test_add_with_no_space(delete, create, resize):
            def _fake_create_image(*args, **kwargs):
                self.called_commands_actual.append('create')
                return self.location

            def _fake_delete_image(
                    target_pool, image_name, snapshot_name=None):
                self.assertEqual(self.location.pool, target_pool)
                self.assertEqual(self.location.image, image_name)
                self.assertEqual(self.location.snapshot, snapshot_name)
                self.called_commands_actual.append('delete')

            def _fake_resize(*args, **kwargs):
                raise MockRBD.NoSpace()

            create.side_effect = _fake_create_image
            delete.side_effect = _fake_delete_image
            resize.side_effect = _fake_resize

            if self.multistore:
                self.assertRaises(exceptions.StorageFull,
                                  self.store.add,
                                  'fake_image_id', self.data_iter, 0)
            else:
                self.assertRaises(exceptions.StorageFull,
                                  self.store.add,
                                  'fake_image_id', self.data_iter, 0,
                                  self.hash_algo)

            self.called_commands_expected = ['create', 'delete']

        _test_add_with_no_space()

    def _test_add_checksums(self):
        """Test that checksums are calculated correctly."""
        self.store.chunk_size = units.Ki
        image_id = 'fake_image_id'
        file_size = 5 * units.Ki  # 5K
        file_contents = b"*" * file_size
        image_file = io.BytesIO(file_contents)
        expected_checksum = hashlib.md5(file_contents,
                                        usedforsecurity=False).hexdigest()
        expected_multihash = hashlib.sha256(file_contents).hexdigest()

        with mock.patch.object(rbd_store.rbd.Image, 'write'):
            if self.multistore:
                loc, size, checksum, metadata = self.store.add(
                    image_id, image_file, file_size)
                self.assertEqual(self.backend, metadata['store'])
            else:
                loc, size, checksum, multihash, _ = self.store.add(
                    image_id, image_file, file_size, self.hash_algo)
                self.assertEqual(expected_multihash, multihash)

        self.assertEqual(expected_checksum, checksum)

    def _test_add_w_image_size_zero_less_resizes(self):
        """Test that correct size is returned with fewer resizes."""
        data_len = 57 * units.Mi
        data_iter = test_utils.FakeData(data_len)
        with mock.patch.object(rbd_store.rbd.Image, 'resize') as resize:
            with mock.patch.object(rbd_store.rbd.Image, 'write') as write:
                if self.multistore:
                    ret = self.store.add('fake_image_id', data_iter, 0)
                else:
                    ret = self.store.add('fake_image_id', data_iter, 0,
                                         self.hash_algo)

                # We expect to trim at the end so +1
                expected = 1
                expected_calls = []
                data_len_temp = data_len
                resize_amount = self.store.WRITE_CHUNKSIZE
                while data_len_temp > 0:
                    resize_amount *= 2
                    expected_calls.append(resize_amount + (data_len -
                                                           data_len_temp))
                    data_len_temp -= resize_amount
                    expected += 1
                self.assertEqual(expected, resize.call_count)
                resize.assert_has_calls([mock.call(call) for call in
                                         expected_calls])
                expected = ([self.store.WRITE_CHUNKSIZE for i in range(int(
                            data_len / self.store.WRITE_CHUNKSIZE))] +
                            [(data_len % self.store.WRITE_CHUNKSIZE)])
                actual = ([len(args[0]) for args, kwargs in
                          write.call_args_list])
                self.assertEqual(expected, actual)
                self.assertEqual(data_len,
                                 resize.call_args_list[-1][0][0])
                self.assertEqual(data_len, ret[1])

    def _test_resize_on_write_ceiling(self):
        """Test resize on write ceiling functionality."""
        image = mock.MagicMock()

        # Non-zero image size means no resize
        ret = self.store._resize_on_write(image, 32, 16, 16)
        self.assertEqual(0, ret)
        image.resize.assert_not_called()

        # Current size is smaller than we need
        self.store.size = 8
        ret = self.store._resize_on_write(image, 0, 16, 16)
        self.assertEqual(8 + self.store.WRITE_CHUNKSIZE * 2, ret)
        self.assertEqual(self.store.WRITE_CHUNKSIZE * 2,
                         self.store.resize_amount)
        image.resize.assert_called_once_with(ret)

        # More reads under the limit do not require a resize
        image.resize.reset_mock()
        self.store.size = ret
        ret = self.store._resize_on_write(image, 0, 64, 16)
        self.assertEqual(8 + self.store.WRITE_CHUNKSIZE * 2, ret)
        image.resize.assert_not_called()

        # Read past the limit triggers another resize
        ret = self.store._resize_on_write(image, 0, ret + 1, 16)
        self.assertEqual(8 + self.store.WRITE_CHUNKSIZE * 6, ret)
        image.resize.assert_called_once_with(ret)
        self.assertEqual(self.store.WRITE_CHUNKSIZE * 4,
                         self.store.resize_amount)

        # Check that we do not resize past the 8G ceiling.

        # Start with resize_amount at 2G, 1G read so far
        image.resize.reset_mock()
        self.store.resize_amount = 2 * units.Gi
        self.store.size = 1 * units.Gi

        # First resize happens and we get to 5G,
        # resize_amount goes to limit of 4G
        ret = self.store._resize_on_write(image, 0, 4097 * units.Mi, 16)
        self.assertEqual(4 * units.Gi, self.store.resize_amount)
        self.assertEqual((1 + 4) * units.Gi, ret)
        self.store.size = ret

        # Second resize happens and we stay at 13, no resize
        # resize amount stays at limit of 8G
        ret = self.store._resize_on_write(image, 0, 6144 * units.Mi, 16)
        self.assertEqual(8 * units.Gi, self.store.resize_amount)
        self.assertEqual((1 + 4 + 8) * units.Gi, ret)
        self.store.size = ret

        # Third resize happens and we get to 21,
        # resize amount stays at limit of 8G
        ret = self.store._resize_on_write(image, 0, 14336 * units.Mi, 16)
        self.assertEqual(8 * units.Gi, self.store.resize_amount)
        self.assertEqual((1 + 4 + 8 + 8) * units.Gi, ret)
        self.store.size = ret

        # Fourth resize happens and we get to 29,
        # resize amount stays at limit of 8G
        ret = self.store._resize_on_write(image, 0, 22528 * units.Mi, 16)
        self.assertEqual(8 * units.Gi, self.store.resize_amount)
        self.assertEqual((1 + 4 + 8 + 8 + 8) * units.Gi, ret)

        image.resize.assert_has_calls([
            mock.call(5 * units.Gi),
            mock.call(13 * units.Gi),
            mock.call(21 * units.Gi),
            mock.call(29 * units.Gi)])

    def _test_create_image_in_native_thread(self):
        """Test creating image in native thread."""
        @mock.patch('oslo_utils.eventletutils.is_monkey_patched')
        def _test_native_thread(mock_patched):
            mock_patched.return_value = True
            # Tests that we use non-0 features from ceph.conf and cast to int.
            fsid = 'fake'
            features = '3'
            conf_get_mock = mock.Mock(return_value=features)
            conn = mock.Mock(conf_get=conf_get_mock)
            ioctxt = mock.sentinel.ioctxt
            name = '1'
            size = 1024
            order = 3
            fake_proxy = mock.MagicMock()
            fake_rbd = mock.MagicMock()

            with mock.patch.object(rbd_store.tpool, 'Proxy') as tpool_mock, \
                    mock.patch.object(rbd_store.rbd, 'RBD') as rbd_mock:
                tpool_mock.return_value = fake_proxy
                rbd_mock.return_value = fake_rbd
                location = self.store._create_image(
                    fsid, conn, ioctxt, name, size, order)
                self.assertEqual(fsid, location.specs['fsid'])
                self.assertEqual(rbd_store.DEFAULT_POOL,
                                 location.specs['pool'])
                self.assertEqual(name, location.specs['image'])
                self.assertEqual(rbd_store.DEFAULT_SNAPNAME,
                                 location.specs['snapshot'])

            tpool_mock.assert_called_once_with(fake_rbd)
            fake_proxy.create.assert_called_once_with(
                ioctxt, name, size, order, old_format=False, features=3)

        _test_native_thread()

    def _test_delete_image_in_native_thread(self):
        """Test deleting image in native thread."""
        @mock.patch('oslo_utils.eventletutils.is_monkey_patched')
        def _test_native_thread(mock_patched):
            mock_patched.return_value = True
            fake_proxy = mock.MagicMock()
            fake_rbd = mock.MagicMock()
            fake_ioctx = mock.MagicMock()

            with mock.patch.object(rbd_store.tpool, 'Proxy') as tpool_mock, \
                    mock.patch.object(rbd_store.rbd, 'RBD') as rbd_mock, \
                    mock.patch.object(
                        self.store, 'get_connection') as mock_conn:

                mock_get_conn = mock_conn.return_value.__enter__.return_value
                mock_ioctx = mock_get_conn.open_ioctx.return_value.__enter__
                mock_ioctx.return_value = fake_ioctx
                tpool_mock.return_value = fake_proxy
                rbd_mock.return_value = fake_rbd

                self.store._delete_image('fake_pool', self.location.image)

                tpool_mock.assert_called_once_with(fake_rbd)
                fake_proxy.remove.assert_called_once_with(fake_ioctx,
                                                          self.location.image)

        _test_native_thread()

    def _test_rbd_proxy(self):
        """Test RBD proxy functionality."""
        @mock.patch.object(rbd_store, 'rbd')
        @mock.patch.object(rbd_store, 'tpool')
        @mock.patch('oslo_utils.eventletutils.is_monkey_patched')
        def _test_proxy(mock_patched, mock_tpool, mock_rbd):
            mock_patched.return_value = False
            self.assertEqual(mock_rbd.RBD(), self.store.RBDProxy())

            mock_patched.return_value = True
            self.assertEqual(
                mock_tpool.Proxy.return_value, self.store.RBDProxy())

        _test_proxy()

    def _test_delete_image(self):
        """Test deleting an image."""
        def _fake_remove(*args, **kwargs):
            self.called_commands_actual.append('remove')

        with mock.patch.object(MockRBD.RBD, 'remove') as remove_image:
            remove_image.side_effect = _fake_remove

            self.store._delete_image('fake_pool', self.location.image)
            self.called_commands_expected = ['remove']

    def _test_delete_image_with_snap_exc_image_has_snap_2(self):
        """Test deleting an image with snapshot that has snapshots (case 2)."""
        def _fake_remove(*args, **kwargs):
            self.called_commands_actual.append('remove')
            raise MockRBD.ImageHasSnapshots()

        mock.patch.object(MockRBD.RBD, 'trash_move',
                          side_effect=MockRBD.ImageBusy).start()

        with mock.patch.object(MockRBD.RBD, 'remove') as remove:
            remove.side_effect = _fake_remove
            self.assertRaises(exceptions.InUseByStore,
                              self.store._delete_image,
                              'fake_pool',
                              self.location.image)

            self.called_commands_expected = ['remove']

        MockRBD.RBD.trash_move.assert_called_once_with(
            mock.ANY, 'fake_image')



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests the Multiple S3 backend store"""

import hashlib
import io
from unittest import mock
import uuid

import boto3
import botocore
from botocore import stub
from oslo_config import cfg

import glance_store as store
from glance_store._drivers import s3
from glance_store import location
from glance_store.tests import base
from glance_store.tests.unit import test_s3_store_base
from glance_store.tests.unit import test_store_capabilities

S3_CONF = {
    's3_store_access_key': 'user',
    's3_store_secret_key': 'key',
    's3_store_host': 'https://s3-region1.com',
    's3_store_bucket': 'glance',
    's3_store_large_object_size': 9,  # over 9MB is large
    's3_store_large_object_chunk_size': 6,  # part size is 6MB
    's3_store_enable_data_integrity_protection': False,
    's3_store_request_checksum_calculation': 'when_required',
    's3_store_response_checksum_validation': 'when_required',
}


class TestMultiS3Store(base.MultiStoreBaseTest,
                       test_s3_store_base.TestS3StoreBase,
                       test_store_capabilities.TestStoreCapabilitiesChecking):
    # NOTE(flaper87): temporary until we
    # can move to a fully-local lib.
    # (Swift store's fault)
    _CONF = cfg.ConfigOpts()

    def setUp(self):
        """Establish a clean test environment."""
        super(TestMultiS3Store, self).setUp()
        enabled_backends = {
            "s3_region1": "s3",
            "s3_region2": "s3"
        }
        self.hash_algo = 'sha256'
        self.conf = self._CONF
        self.conf(args=[])
        self.conf.register_opt(cfg.DictOpt('enabled_backends'))
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(self.conf)
        self.config(default_backend='s3_region1', group='glance_store')

        # set s3 related config options
        self.config(group='s3_region1',
                    s3_store_access_key='user',
                    s3_store_secret_key='key',
                    s3_store_host='https://s3-region1.com',
                    s3_store_region_name='custom_region_name',
                    s3_store_cacert='path/to/cert/bundle.pem',
                    s3_store_bucket='glance',
                    s3_store_large_object_size=S3_CONF[
                        's3_store_large_object_size'
                    ],
                    s3_store_large_object_chunk_size=6,
                    s3_store_enable_data_integrity_protection=False,
                    s3_store_request_checksum_calculation='when_required',
                    s3_store_response_checksum_validation='when_required')

        self.config(group='s3_region2',
                    s3_store_access_key='user',
                    s3_store_secret_key='key',
                    s3_store_host='http://s3-region2.com',
                    s3_store_bucket='glance',
                    s3_store_large_object_size=S3_CONF[
                        's3_store_large_object_size'
                    ],
                    s3_store_large_object_chunk_size=6,
                    s3_store_enable_data_integrity_protection=False,
                    s3_store_request_checksum_calculation='when_required',
                    s3_store_response_checksum_validation='when_required')
        # Ensure stores + locations cleared
        location.SCHEME_TO_CLS_BACKEND_MAP = {}
        store.create_multi_stores(self.conf)

        self.addCleanup(setattr, location, 'SCHEME_TO_CLS_BACKEND_MAP',
                        dict())
        self.addCleanup(self.conf.reset)

        self.store = s3.Store(self.conf, backend="s3_region1")
        self.store.configure()
        self.register_store_backend_schemes(self.store, 's3', 's3_region1')

        # Set default multistore and backend parameters
        self.multistore = True
        self.backend = 's3_region1'

    def test_location_url_prefix_is_set(self):
        expected_url_prefix = "s3+https://user:key@s3-region1.com/glance"
        self.assertEqual(expected_url_prefix, self.store.url_prefix)

    def test_get_invalid_bucket_name(self):
        self._test_get_invalid_bucket_name()

    def test_client_custom_region_name(self):
        self._test_client_custom_region_name()

    def test_client_custom_ca_cert_bundle(self):
        self._test_client_custom_ca_cert_bundle()

    def test_get(self):
        self._test_get()

    def test_partial_get(self):
        self._test_partial_get()

    def test_get_non_existing(self):
        self._test_get_non_existing()

    def test_add_singlepart(self):
        """Test that we can add an image via the s3 backend."""
        self._test_add_singlepart()

    def test_add_singlepart_size_exceeding_max_size(self):
        self._test_add_singlepart_size_exceeding_max_size()

    def test_add_singlepart_write_less_than_declared(self):
        self._test_add_singlepart_write_less_than_declared()

    def test_add_singlepart_bigger_than_write_chunk(self):
        self._test_add_singlepart_bigger_than_write_chunk()

    def test_add_different_backend(self):
        self._test_add_different_backend()

    def test_add_with_verifier(self):
        self._test_add_with_verifier()

    def test_add_multipart(self):
        self._test_add_multipart()

    def test_add_multipart_size_exceeding_max_size(self):
        self._test_add_multipart_size_exceeding_max_size()

    def test_add_multipart_write_less_than_declared(self):
        self._test_add_multipart_write_less_than_declared()

    def test_add_already_existing(self):
        self._test_add_already_existing()

    def test_delete_non_existing(self):
        self._test_delete_non_existing()

    def test_no_access_key(self):
        self._test_no_access_key()

    def test_no_secret_key(self):
        self._test_no_secret_key()

    def test_no_host(self):
        self._test_no_host()

    def test_no_bucket(self):
        self._test_no_bucket()

    def test_get_s3_good_location(self):
        self._test_get_s3_good_location()

    def test_add_multipart_zero_byte_image(self):
        """Bug #2124829: add a zero-byte image using multipart upload."""
        expected_image_id = str(uuid.uuid4())
        expected_s3_size = 0
        expected_s3_contents = b""
        expected_checksum = hashlib.md5(expected_s3_contents,
                                        usedforsecurity=False).hexdigest()
        expected_multihash = hashlib.sha256(expected_s3_contents).hexdigest()
        backend_conf = getattr(self.conf, self.backend)

        expected_location = test_s3_store_base.format_s3_location(
            backend_conf.s3_store_access_key,
            backend_conf.s3_store_secret_key,
            backend_conf.s3_store_host,
            backend_conf.s3_store_bucket,
            expected_image_id)
        image_s3 = io.BytesIO(expected_s3_contents)

        self.config(group=self.backend, s3_store_large_object_size=0)
        self.store.configure()
        fake_s3_client = botocore.session.get_session().create_client('s3')

        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_response(method='head_bucket',
                                 service_response={},
                                 expected_params={
                                     'Bucket': backend_conf.s3_store_bucket
                                 })
            stubber.add_client_error(
                method='head_object', service_error_code='404',
                service_message='',
                expected_params={
                    'Bucket': backend_conf.s3_store_bucket,
                    'Key': expected_image_id})
            stubber.add_response(method='create_multipart_upload',
                                 service_response={
                                     "Bucket": backend_conf.s3_store_bucket,
                                     "Key": expected_image_id,
                                     "UploadId": 'UploadId'
                                 },
                                 expected_params={
                                     "Bucket": backend_conf.s3_store_bucket,
                                     "Key": expected_image_id,
                                 })

            stubber.add_response(method='abort_multipart_upload',
                                 service_response={},
                                 expected_params={
                                     'Bucket': backend_conf.s3_store_bucket,
                                     'Key': expected_image_id,
                                     'UploadId': 'UploadId'
                                 })
            stubber.add_response(method='put_object',
                                 service_response={},
                                 expected_params={
                                     'Bucket': backend_conf.s3_store_bucket,
                                     'Key': expected_image_id,
                                     'Body': b''
                                 })

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                mock_client.return_value = fake_s3_client

                loc, size, checksum, multihash, metadata = \
                    self.store.add(expected_image_id, image_s3,
                                   expected_s3_size,
                                   self.hash_algo)

                stubber.assert_no_pending_responses()
                self.assertEqual(self.backend, metadata["store"])
                self.assertEqual(expected_location, loc)
                self.assertEqual(expected_s3_size, size)
                self.assertEqual(expected_checksum, checksum)
                self.assertEqual(expected_multihash, multihash)

    def test_get_my_object_storage_location(self):
        self._test_get_my_object_storage_location()

    def test_config_with_data_integrity_protection_disabled(self):
        self._test_config_with_data_integrity_protection_disabled()

    def test_config_with_data_integrity_protection_enabled(self):
        self._test_config_with_data_integrity_protection_enabled()

    def test_config_fallback_for_old_boto3(self):
        self._test_config_fallback_for_old_boto3()



# Copyright 2021 RedHat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from cinderclient.apiclient import exceptions as apiclient_exception
from cinderclient import exceptions as cinder_exception
from keystoneauth1 import exceptions as keystone_exc
from oslo_utils import excutils
import retrying

from glance_store import exceptions
from glance_store.i18n import _LE

LOG = logging.getLogger(__name__)


def handle_exceptions(method):
    """Transforms the exception for the volume but keeps its traceback intact.
    """
    def wrapper(self, ctx, volume_id, *args, **kwargs):
        try:
            res = method(self, ctx, volume_id, *args, **kwargs)
        except (keystone_exc.NotFound,
                cinder_exception.NotFound,
                cinder_exception.OverLimit) as e:
            raise exceptions.BackendException(str(e))
        return res
    return wrapper


def _retry_on_internal_server_error(e):
    if isinstance(e, apiclient_exception.InternalServerError):
        return True
    return False


def _retry_on_bad_request(e):
    if isinstance(e, cinder_exception.BadRequest):
        return True
    return False


class API(object):
    """API for interacting with the cinder."""

    @handle_exceptions
    def create(self, client, size, name,
               volume_type=None, metadata=None):

        kwargs = dict(volume_type=volume_type,
                      metadata=metadata,
                      name=name)

        volume = client.volumes.create(size, **kwargs)
        return volume

    @handle_exceptions
    def get(self, client, volume_id):
        """Get the details about a volume given it's ID.

        :param client: cinderclient object
        :param volume_id: UUID of the volume to get
        """
        return client.volumes.get(volume_id)

    def delete(self, client, volume_id):
        client.volumes.delete(volume_id)

    @retrying.retry(stop_max_attempt_number=5,
                    retry_on_exception=_retry_on_bad_request,
                    wait_exponential_multiplier=1000,
                    wait_exponential_max=10000)
    @handle_exceptions
    def attachment_create(self, client, volume_id, connector=None,
                          mountpoint=None, mode=None):
        """Create a volume attachment. This requires microversion >= 3.54.

        The attachment_create call was introduced in microversion 3.27. We
        need 3.54 as minimum here as we need attachment_complete to finish the
        attaching process and it which was introduced in version 3.44 and
        we also pass the attach mode which was introduced in version 3.54.

        :param client: cinderclient object
        :param volume_id: UUID of the volume on which to create the attachment.
        :param connector: host connector dict; if None, the attachment will
            be 'reserved' but not yet attached.
        :param mountpoint: Optional mount device name for the attachment,
            e.g. "/dev/vdb". This is only used if a connector is provided.
        :param mode: The mode in which the attachment is made i.e.
            read only(ro) or read/write(rw)
        :returns: a dict created from the
            cinderclient.v3.attachments.VolumeAttachment object with a backward
            compatible connection_info dict
        """
        if connector and mountpoint and 'mountpoint' not in connector:
            connector['mountpoint'] = mountpoint

        try:
            attachment_ref = client.attachments.create(
                volume_id, connector, mode=mode)
            return attachment_ref
        except cinder_exception.ClientException as ex:
            with excutils.save_and_reraise_exception():
                # While handling simultaneous requests, the volume can be
                # in different states and we retry on attachment_create
                # until the volume reaches a valid state for attachment.
                # Hence, it is better to not log 400 cases as no action
                # from users is needed in this case
                if getattr(ex, 'code', None) != 400:
                    LOG.error(_LE('Create attachment failed for volume '
                                  '%(volume_id)s. Error: %(msg)s '
                                  'Code: %(code)s'),
                              {'volume_id': volume_id,
                               'msg': str(ex),
                               'code': getattr(ex, 'code', None)})

    @handle_exceptions
    def attachment_get(self, client, attachment_id):
        """Gets a volume attachment.

        :param client: cinderclient object
        :param attachment_id: UUID of the volume attachment to get.
        :returns: a dict created from the
            cinderclient.v3.attachments.VolumeAttachment object with a backward
            compatible connection_info dict
        """
        try:
            attachment_ref = client.attachments.show(
                attachment_id)
            return attachment_ref
        except cinder_exception.ClientException as ex:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE('Show attachment failed for attachment '
                              '%(id)s. Error: %(msg)s Code: %(code)s'),
                          {'id': attachment_id,
                           'msg': str(ex),
                           'code': getattr(ex, 'code', None)})

    @handle_exceptions
    def attachment_update(self, client, attachment_id, connector,
                          mountpoint=None):
        """Updates the connector on the volume attachment. An attachment
        without a connector is considered reserved but not fully attached.

        :param client: cinderclient object
        :param attachment_id: UUID of the volume attachment to update.
        :param connector: host connector dict. This is required when updating
            a volume attachment. To terminate a connection, the volume
            attachment for that connection must be deleted.
        :param mountpoint: Optional mount device name for the attachment,
            e.g. "/dev/vdb". Theoretically this is optional per volume backend,
            but in practice it's normally required so it's best to always
            provide a value.
        :returns: a dict created from the
            cinderclient.v3.attachments.VolumeAttachment object with a backward
            compatible connection_info dict
        """
        if mountpoint and 'mountpoint' not in connector:
            connector['mountpoint'] = mountpoint

        try:
            attachment_ref = client.attachments.update(
                attachment_id, connector)
            return attachment_ref
        except cinder_exception.ClientException as ex:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE('Update attachment failed for attachment '
                              '%(id)s. Error: %(msg)s Code: %(code)s'),
                          {'id': attachment_id,
                           'msg': str(ex),
                           'code': getattr(ex, 'code', None)})

    @handle_exceptions
    def attachment_complete(self, client, attachment_id):
        """Marks a volume attachment complete.

        This call should be used to inform Cinder that a volume attachment is
        fully connected on the host so Cinder can apply the necessary state
        changes to the volume info in its database.

        :param client: cinderclient object
        :param attachment_id: UUID of the volume attachment to update.
        """
        try:
            client.attachments.complete(attachment_id)
        except cinder_exception.ClientException as ex:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE('Complete attachment failed for attachment '
                              '%(id)s. Error: %(msg)s Code: %(code)s'),
                          {'id': attachment_id,
                           'msg': str(ex),
                           'code': getattr(ex, 'code', None)})

    @handle_exceptions
    @retrying.retry(stop_max_attempt_number=5,
                    retry_on_exception=_retry_on_internal_server_error)
    def attachment_delete(self, client, attachment_id):
        try:
            client.attachments.delete(attachment_id)
        except cinder_exception.ClientException as ex:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE('Delete attachment failed for attachment '
                              '%(id)s. Error: %(msg)s Code: %(code)s'),
                          {'id': attachment_id,
                           'msg': str(ex),
                           'code': getattr(ex, 'code', None)})

    @handle_exceptions
    def extend_volume(self, client, volume, new_size):
        """Extend volume

        :param client: cinderclient object
        :param volume: UUID of the volume to extend
        :param new_size: new size of the volume after extend
        """
        try:
            client.volumes.extend(volume, new_size)
        except cinder_exception.ClientException as ex:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE('Extend volume failed for volume '
                              '%(id)s. Error: %(msg)s Code: %(code)s'),
                          {'id': volume.id,
                           'msg': str(ex),
                           'code': getattr(ex, 'code', None)})



# Copyright (c) 2014 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Glance Store exception subclasses"""

import urllib.parse

from glance_store.i18n import _


class BackendException(Exception):
    pass


class UnsupportedBackend(BackendException):
    pass


class RedirectException(Exception):
    def __init__(self, url):
        self.url = urllib.parse.urlparse(url)


class GlanceStoreException(Exception):
    """
    Base Glance Store Exception

    To correctly use this class, inherit from it and define
    a 'message' property. That message will get printf'd
    with the keyword arguments provided to the constructor.
    """
    message = _("An unknown exception occurred")

    def __init__(self, message=None, **kwargs):
        if not message:
            message = self.message
        try:
            if kwargs:
                message = message % kwargs
        except Exception:
            pass
        self.msg = message
        super(GlanceStoreException, self).__init__(message)


class MissingCredentialError(GlanceStoreException):
    message = _("Missing required credential: %(required)s")


class BadAuthStrategy(GlanceStoreException):
    message = _("Incorrect auth strategy, expected \"%(expected)s\" but "
                "received \"%(received)s\"")


class AuthorizationRedirect(GlanceStoreException):
    message = _("Redirecting to %(uri)s for authorization.")


class NotFound(GlanceStoreException):
    message = _("Image %(image)s not found")


class UnknownHashingAlgo(GlanceStoreException):
    message = _("Unknown hashing algorithm identifier: %(algo)s")


class UnknownScheme(GlanceStoreException):
    message = _("Unknown scheme '%(scheme)s' found in URI")


class BadStoreUri(GlanceStoreException):
    message = _("The Store URI was malformed: %(uri)s")


class Duplicate(GlanceStoreException):
    message = _("Image %(image)s already exists")


class StorageFull(GlanceStoreException):
    message = _("There is not enough disk space on the image storage media.")


class StorageWriteDenied(GlanceStoreException):
    message = _("Permission to write image storage media denied.")


class AuthBadRequest(GlanceStoreException):
    message = _("Connect error/bad request to Auth service at URL %(url)s.")


class AuthUrlNotFound(GlanceStoreException):
    message = _("Auth service at URL %(url)s not found.")


class AuthorizationFailure(GlanceStoreException):
    message = _("Authorization failed.")


class NotAuthenticated(GlanceStoreException):
    message = _("You are not authenticated.")


class Forbidden(GlanceStoreException):
    message = _("You are not authorized to complete this action.")


class Invalid(GlanceStoreException):
    # NOTE(NiallBunting) This could be deprecated however the debtcollector
    # seems to have problems deprecating this as well as the subclasses.
    message = _("Data supplied was not valid.")


class BadStoreConfiguration(GlanceStoreException):
    message = _("Store %(store_name)s could not be configured correctly. "
                "Reason: %(reason)s")


class DriverLoadFailure(GlanceStoreException):
    message = _("Driver %(driver_name)s could not be loaded.")


class StoreDeleteNotSupported(GlanceStoreException):
    message = _("Deleting images from this store is not supported.")


class StoreGetNotSupported(GlanceStoreException):
    message = _("Getting images from this store is not supported.")


class StoreRandomGetNotSupported(StoreGetNotSupported):
    message = _("Getting images randomly from this store is not supported. "
                "Offset: %(offset)s, length: %(chunk_size)s")


class StoreAddDisabled(GlanceStoreException):
    message = _("Configuration for store failed. Adding images to this "
                "store is disabled.")


class MaxRedirectsExceeded(GlanceStoreException):
    message = _("Maximum redirects (%(redirects)s) was exceeded.")


class NoServiceEndpoint(GlanceStoreException):
    message = _("Response from Keystone does not contain a Glance endpoint.")


class RegionAmbiguity(GlanceStoreException):
    message = _("Multiple 'image' service matches for region %(region)s. This "
                "generally means that a region is required and you have not "
                "supplied one.")


class RemoteServiceUnavailable(GlanceStoreException):
    message = _("Remote server where the image is present is unavailable.")


class HasSnapshot(GlanceStoreException):
    message = _("The image cannot be deleted because it has snapshot(s).")


class InUseByStore(GlanceStoreException):
    message = _("The image cannot be deleted because it is in use through "
                "the backend store outside of Glance.")


class HostNotInitialized(GlanceStoreException):
    message = _("The glance cinder store host %(host)s which will used to "
                "perform nfs mount/umount operations isn't initialized.")


class TimeoutError(GlanceStoreException):
    message = _("Operation timed out after %(timeout)ds. Check storage "
                "health and connectivity.")



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import configparser
import logging
import warnings

from oslo_config import cfg

from glance_store import driver
from glance_store import exceptions
from glance_store.i18n import _, _LE

swift_opts = [
    cfg.StrOpt('default_swift_reference',
               default="ref1",
               help="""
Reference to default Swift account/backing store parameters.

Provide a string value representing a reference to the default set
of parameters required for using swift account/backing store for
image storage. The default reference value for this configuration
option is 'ref1'. This configuration option dereferences the
parameters and facilitates image storage in Swift storage backend
every time a new image is added.

Possible values:
    * A valid string value

Related options:
    * None

"""),
    cfg.StrOpt('swift_store_auth_version', default='3',
               choices=['3'],
               help='The authentication version to be used. Currently '
                    'The only valid version is 3.',
               deprecated_for_removal=True,
               deprecated_reason="""
This option is kept for backword-compatibility reasons but is no longer
required, because only the single version (3) is supported now.
"""),
    cfg.StrOpt('swift_store_auth_address',
               help='The address where the Swift authentication '
                    'service is listening.',
               deprecated_for_removal=True,
               deprecated_reason="""
The option 'auth_address' in the Swift back-end configuration file is
used instead.
"""),
    cfg.StrOpt('swift_store_user', secret=True,
               help='The user to authenticate against the Swift '
                    'authentication service.',
               deprecated_for_removal=True,
               deprecated_reason="""
The option 'user' in the Swift back-end configuration file is set instead.
"""),
    cfg.StrOpt('swift_store_key', secret=True,
               help='Auth key for the user authenticating against the '
                    'Swift authentication service.',
               deprecated_for_removal=True,
               deprecated_reason="""
The option 'key' in the Swift back-end configuration file is used
to set the authentication key instead.
"""),
    cfg.StrOpt('swift_store_application_credential_id', secret=True,
               help="""
Application credential ID for authenticating against Swift.

This option specifies the application credential ID to use for
authenticating with the Swift backend. When set along with
swift_store_application_credential_secret, the Swift driver will
use V3ApplicationCredential authentication instead of password
authentication.

This enables Zero Downtime Password Rotation (ZDPR) support for
Swift backend operations, as application credentials are not
affected by password rotation.

If not set, the driver falls back to password authentication
using swift_store_user and swift_store_key.

Possible values:
    * A valid application credential ID string

Related options:
    * swift_store_application_credential_secret

"""),
    cfg.StrOpt('swift_store_application_credential_secret', secret=True,
               help="""
Application credential secret for authenticating against Swift.

This option specifies the application credential secret to use
for authenticating with the Swift backend. When set along with
swift_store_application_credential_id, the Swift driver will
use V3ApplicationCredential authentication instead of password
authentication.

This enables Zero Downtime Password Rotation (ZDPR) support for
Swift backend operations, as application credentials are not
affected by password rotation.

If not set, the driver falls back to password authentication
using swift_store_user and swift_store_key.

Possible values:
    * A valid application credential secret string

Related options:
    * swift_store_application_credential_id

"""),
    cfg.StrOpt('swift_store_project_name',
               help='Project name for authenticating with application '
                    'credentials against the Swift authentication service.',
               deprecated_for_removal=True,
               deprecated_reason="""
The option 'project_name' in the Swift back-end configuration file is
used instead.
"""),
    cfg.StrOpt('swift_store_project_id',
               help='Project ID for authenticating with application '
                    'credentials against the Swift authentication service.',
               deprecated_for_removal=True,
               deprecated_reason="""
The option 'project_id' in the Swift back-end configuration file is
used instead.
"""),
    cfg.StrOpt('swift_store_config_file',
               default=None,
               help="""
Absolute path to the file containing the swift account(s)
configurations.

Include a string value representing the path to a configuration
file that has references for each of the configured Swift
account(s)/backing stores. By default, no file path is specified
and customized Swift referencing is disabled. Configuring this
option is highly recommended while using Swift storage backend for
image storage as it avoids storage of credentials in the database.

NOTE: Please do not configure this option if you have set
``swift_store_multi_tenant`` to ``True``.

Possible values:
    * String value representing an absolute path on the glance-api
      node

Related options:
    * swift_store_multi_tenant

"""),
]


class SwiftConfigParser(configparser.ConfigParser):

    def get(self, *args, **kwargs):
        value = super(configparser.ConfigParser, self).get(*args, **kwargs)
        return self._process_quotes(value)

    @staticmethod
    def _process_quotes(value):
        if value:
            if value[0] in "\"'":
                if len(value) == 1 or value[-1] != value[0]:
                    raise ValueError('Non-closed quote: %s' %
                                     value)
                value = value[1:-1]
        return value


CONFIG = SwiftConfigParser()

LOG = logging.getLogger(__name__)


def is_multiple_swift_store_accounts_enabled(conf, backend=None):
    if backend:
        cfg_file = getattr(conf, backend).swift_store_config_file
    else:
        cfg_file = conf.glance_store.swift_store_config_file

    if cfg_file is None:
        return False
    return True


class SwiftParams(object):
    def __init__(self, conf, backend=None):
        self.conf = conf
        self.backend_group = backend
        if is_multiple_swift_store_accounts_enabled(
                self.conf, backend=backend):
            self.params = self._load_config()
        else:
            self.params = self._form_default_params()

    def _form_default_params(self):
        default = {}
        from glance_store._drivers.swift import store as swift_store
        store_opts = swift_store._SWIFT_OPTS + swift_opts
        if self.backend_group:
            glance_store = driver.BackendGroupConfiguration(
                store_opts, self.backend_group, conf=self.conf)
        else:
            glance_store = self.conf.glance_store
        if (
                glance_store.swift_store_application_credential_id and
                glance_store.swift_store_application_credential_secret and
                glance_store.swift_store_auth_address
        ):
            default['application_credential_id'] = (
                glance_store.swift_store_application_credential_id)
            default['application_credential_secret'] = (
                glance_store.swift_store_application_credential_secret)
            default['auth_address'] = glance_store.swift_store_auth_address
            default['project_domain_id'] = 'default'
            default['project_domain_name'] = None
            default['user_domain_id'] = 'default'
            default['user_domain_name'] = None
            default['auth_version'] = glance_store.swift_store_auth_version
            if hasattr(glance_store, 'swift_store_project_name'):
                default['project_name'] = glance_store.swift_store_project_name
            if hasattr(glance_store, 'swift_store_project_id'):
                default['project_id'] = glance_store.swift_store_project_id
            return {glance_store.default_swift_reference: default}
        elif (
                glance_store.swift_store_user and
                glance_store.swift_store_key and
                glance_store.swift_store_auth_address
        ):

            default['user'] = glance_store.swift_store_user
            default['key'] = glance_store.swift_store_key
            default['auth_address'] = glance_store.swift_store_auth_address
            default['project_domain_id'] = 'default'
            default['project_domain_name'] = None
            default['user_domain_id'] = 'default'
            default['user_domain_name'] = None
            default['auth_version'] = glance_store.swift_store_auth_version
            return {glance_store.default_swift_reference: default}
        return {}

    def _load_config(self):
        if self.backend_group:
            scf = getattr(self.conf,
                          self.backend_group).swift_store_config_file
        else:
            scf = self.conf.glance_store.swift_store_config_file
        try:
            conf_file = self.conf.find_file(scf)
            CONFIG.read(conf_file)
        except Exception as e:
            msg = (_("swift config file "
                     "%(conf)s:%(exc)s not found"),
                   {'conf': scf,
                    'exc': e})
            LOG.error(msg)
            raise exceptions.BadStoreConfiguration(store_name='swift',
                                                   reason=msg)
        account_params = {}
        account_references = CONFIG.sections()

        for ref in account_references:
            reference = {}
            try:
                try:
                    reference['application_credential_id'] = CONFIG.get(
                        ref, 'application_credential_id')
                    reference['application_credential_secret'] = CONFIG.get(
                        ref, 'application_credential_secret')
                    reference['auth_address'] = CONFIG.get(ref, 'auth_address')
                    reference['project_name'] = CONFIG.get(
                        ref, 'project_name', fallback=None)
                    reference['project_id'] = CONFIG.get(
                        ref, 'project_id', fallback=None)
                except configparser.NoOptionError:
                    reference['application_credential_id'] = None
                    reference['application_credential_secret'] = None
                    for param in ('auth_address', 'user', 'key'):
                        reference[param] = CONFIG.get(ref, param)

                reference['project_domain_name'] = CONFIG.get(
                    ref, 'project_domain_name', fallback=None)
                reference['project_domain_id'] = CONFIG.get(
                    ref, 'project_domain_id', fallback=None)
                if (reference['project_domain_name'] is None and
                        reference['project_domain_id'] is None):
                    reference['project_domain_id'] = 'default'

                reference['user_domain_name'] = CONFIG.get(
                    ref, 'user_domain_name', fallback=None)
                reference['user_domain_id'] = CONFIG.get(
                    ref, 'user_domain_id', fallback=None)
                if (reference['user_domain_name'] is None and
                        reference['user_domain_id'] is None):
                    reference['user_domain_id'] = 'default'

                try:
                    reference['auth_version'] = CONFIG.get(ref, 'auth_version')
                    warnings.warn(
                        'The auth_version option is deprecated. It is kept '
                        'for backword-compatibility reasons but will be '
                        'removed in a future release.',
                        DeprecationWarning)
                except configparser.NoOptionError:
                    if self.backend_group:
                        conf_group = getattr(self.conf, self.backend_group)
                    else:
                        conf_group = self.conf.glance_store
                    try:
                        av = getattr(conf_group, 'swift_store_auth_version')
                    except AttributeError:
                        av = '3'
                    reference['auth_version'] = av or '3'

                if reference.get('auth_version') != '3':
                    reference['auth_version'] = '3'

                account_params[ref] = reference
            except (ValueError, SyntaxError, configparser.NoOptionError):
                LOG.exception(_LE("Invalid format of swift store config cfg"))
        return account_params



# Copyright 2010 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Storage backend for S3 or Storage Servers that follow the S3 Protocol"""

from concurrent import futures
import io
import logging
import math
import re
import urllib

try:
    from boto3 import session as boto_session
    from botocore import client as boto_client
    from botocore import exceptions as boto_exceptions
    from botocore import utils as boto_utils
except ImportError:
    boto_session = None
    boto_client = None
    boto_exceptions = None
    boto_utils = None

from oslo_config import cfg
from oslo_utils import units

import glance_store
from glance_store import capabilities
from glance_store.common import utils
import glance_store.driver
from glance_store import exceptions
from glance_store.i18n import _
import glance_store.location

LOG = logging.getLogger(__name__)

DEFAULT_LARGE_OBJECT_SIZE = 100          # 100M
DEFAULT_LARGE_OBJECT_CHUNK_SIZE = 10     # 10M
DEFAULT_LARGE_OBJECT_MIN_CHUNK_SIZE = 5  # 5M
DEFAULT_THREAD_POOLS = 10                # 10 pools
MAX_PART_NUM = 10000                     # 10000 upload parts

_S3_OPTS = [
    cfg.StrOpt('s3_store_host',
               help="""
The host where the S3 server is listening.

This configuration option sets the host of the S3 or S3 compatible storage
Server. This option is required when using the S3 storage backend.
The host can contain a DNS name (e.g. s3.amazonaws.com, my-object-storage.com)
or an IP address (127.0.0.1).

Possible values:
    * A valid DNS name
    * A valid IPv4 address

Related Options:
    * s3_store_access_key
    * s3_store_secret_key

"""),
    cfg.StrOpt('s3_store_region_name',
               default='',
               help="""
The S3 region name.

This parameter will set the region_name used by boto.
If this parameter is not set, we we will try to compute it from the
s3_store_host.

Possible values:
    * A valid region name

Related Options:
    * s3_store_host

"""),
    cfg.StrOpt('s3_store_access_key',
               secret=True,
               help="""
The S3 query token access key.

This configuration option takes the access key for authenticating with the
Amazon S3 or S3 compatible storage server. This option is required when using
the S3 storage backend.

Possible values:
    * Any string value that is the access key for a user with appropriate
      privileges

Related Options:
    * s3_store_host
    * s3_store_secret_key

"""),
    cfg.StrOpt('s3_store_secret_key',
               secret=True,
               help="""
The S3 query token secret key.

This configuration option takes the secret key for authenticating with the
Amazon S3 or S3 compatible storage server. This option is required when using
the S3 storage backend.

Possible values:
    * Any string value that is a secret key corresponding to the access key
      specified using the ``s3_store_host`` option

Related Options:
    * s3_store_host
    * s3_store_access_key

"""),
    cfg.StrOpt('s3_store_bucket',
               help="""
The S3 bucket to be used to store the Glance data.

This configuration option specifies where the glance images will be stored
in the S3. If ``s3_store_create_bucket_on_put`` is set to true, it will be
created automatically even if the bucket does not exist.

Possible values:
    * Any string value

Related Options:
    * s3_store_create_bucket_on_put
    * s3_store_bucket_url_format

"""),
    cfg.BoolOpt('s3_store_create_bucket_on_put',
                default=False,
                help="""
Determine whether S3 should create a new bucket.

This configuration option takes boolean value to indicate whether Glance should
create a new bucket to S3 if it does not exist.

Possible values:
    * Any Boolean value

Related Options:
    * None

"""),
    cfg.StrOpt('s3_store_bucket_url_format',
               default='auto',
               help="""
The S3 calling format used to determine the object.

This configuration option takes access model that is used to specify the
address of an object in an S3 bucket.

NOTE:
In ``path``-style, the endpoint for the object looks like
'https://s3.amazonaws.com/bucket/example.img'.
And in ``virtual``-style, the endpoint for the object looks like
'https://bucket.s3.amazonaws.com/example.img'.
If you do not follow the DNS naming convention in the bucket name, you can
get objects in the path style, but not in the virtual style.

Possible values:
    * Any string value of ``auto``, ``virtual``, or ``path``

Related Options:
    * s3_store_bucket

"""),
    cfg.IntOpt('s3_store_large_object_size',
               default=DEFAULT_LARGE_OBJECT_SIZE,
               min=0,
               help="""
What size, in MB, should S3 start chunking image files and do a multipart
upload in S3.

This configuration option takes a threshold in MB to determine whether to
upload the image to S3 as is or to split it (Multipart Upload).

Note: You can only split up to 10,000 images.

Possible values:
    * Any positive integer value or zero

Related Options:
    * s3_store_large_object_chunk_size
    * s3_store_thread_pools

"""),
    cfg.IntOpt('s3_store_large_object_chunk_size',
               default=DEFAULT_LARGE_OBJECT_CHUNK_SIZE,
               help="""
What multipart upload part size, in MB, should S3 use when uploading parts.

This configuration option takes the image split size in MB for Multipart
Upload.

Note: You can only split up to 10,000 images.

Possible values:
    * Any positive integer value (must be greater than or equal to 5M)

Related Options:
    * s3_store_large_object_size
    * s3_store_thread_pools

"""),
    cfg.IntOpt('s3_store_thread_pools',
               default=DEFAULT_THREAD_POOLS,
               help="""
The number of thread pools to perform a multipart upload in S3.

This configuration option takes the number of thread pools when performing a
Multipart Upload.

Possible values:
    * Any positive integer value

Related Options:
    * s3_store_large_object_size
    * s3_store_large_object_chunk_size

"""),
    cfg.StrOpt('s3_store_cacert',
               default='',
               help="""
The path to the CA cert bundle to use. The default value (an empty string)
forces the use of the default CA cert bundle used by botocore.

Possible values:
    * A path to the CA cert bundle to use
    * An empty string to use the default CA cert bundle used by botocore

"""),
    cfg.BoolOpt('s3_store_enable_data_integrity_protection',
                default=False,
                help="""
Turn on S3 data integrity protections by enabling request checksum calculation
and response checksum validation. Setting this to False is the same as setting
the AWS_REQUEST_CHECKSUM_CALCULATION and AWS_RESPONSE_CHECKSUM_VALIDATION
environment variables to "when_required". Setting this to True gives the user
fine-grained control over the data integrity protections by using the
s3_store_request_checksum_calculation and s3_store_response_checksum_validation
options.

This defaults to False as to not change the default behaviour in glance_store,
which prevents upload failures when using S3-compatible storage backends that
do not implement S3 data integrity protection.

Related Options:
    * s3_store_request_checksum_calculation
    * s3_store_response_checksum_validation
"""),
    cfg.StrOpt('s3_store_request_checksum_calculation',
               default='when_required',
               choices=['when_required', 'when_supported'],
               help="""
Controls when checksums are calculated for S3 upload requests.

This configuration option provides fine-grained control over request checksum
calculation behavior. It maps directly to the botocore Config parameter
request_checksum_calculation.

Possible values:
    * when_required - Only calculate checksums when required by the operation
    * when_supported - Calculate checksums when supported by the operation

This option is only used when s3_store_enable_data_integrity_protection is
set to True. When s3_store_enable_data_integrity_protection is False, both
request and response checksums use 'when_required' regardless of this setting.

Related Options:
    * s3_store_response_checksum_validation
    * s3_store_enable_data_integrity_protection

"""),
    cfg.StrOpt('s3_store_response_checksum_validation',
               default='when_required',
               choices=['when_required', 'when_supported'],
               help="""
Controls when checksums are validated for S3 download responses.

This configuration option provides fine-grained control over response checksum
validation behavior. It maps directly to the botocore Config parameter
response_checksum_validation.

Possible values:
    * when_required - Only validate checksums when required by the operation
    * when_supported - Validate checksums when supported by the operation

This option is only used when s3_store_enable_data_integrity_protection is
set to True. When s3_store_enable_data_integrity_protection is False, both
request and response checksums use 'when_required' regardless of this setting.

Related Options:
    * s3_store_request_checksum_calculation
    * s3_store_enable_data_integrity_protection

"""),
]


class UploadPart(object):
    """The class for the upload part."""
    def __init__(self, mpu, fp, partnum, chunks):
        self.mpu = mpu
        self.partnum = partnum
        self.fp = fp
        self.size = 0
        self.chunks = chunks
        self.etag = {}
        self.success = True


def run_upload(s3_client, bucket, key, part):
    """Upload the upload part into S3 and set returned etag and size to its
    part info.

    :param s3_client: An object with credentials to connect to S3
    :param bucket: The S3 bucket name
    :param key: The object name to be stored (image identifier)
    :param part: UploadPart object which used during multipart upload
    """
    pnum = part.partnum
    bsize = part.chunks
    upload_id = part.mpu['UploadId']
    LOG.debug("Uploading upload part in S3 partnum=%(pnum)d, "
              "size=%(bsize)d, key=%(key)s, UploadId=%(UploadId)s",
              {'pnum': pnum, 'bsize': bsize, 'key': key,
               'UploadId': upload_id})

    try:
        key = s3_client.upload_part(Body=part.fp,
                                    Bucket=bucket,
                                    ContentLength=bsize,
                                    Key=key,
                                    PartNumber=pnum,
                                    UploadId=upload_id)
        part.etag[part.partnum] = key['ETag']
        part.size = bsize
    except boto_exceptions.ClientError as e:
        error_code = e.response['Error']['Code']
        error_message = e.response['Error']['Message']
        LOG.warning("Failed to upload part in S3 partnum=%(pnum)d, "
                    "size=%(bsize)d, error code=%(error_code)d, "
                    "error message=%(error_message)s",
                    {'pnum': pnum, 'bsize': bsize, 'error_code': error_code,
                     'error_message': error_message})
        part.success = False
    finally:
        part.fp.close()


class StoreLocation(glance_store.location.StoreLocation):
    """Class describing an S3 URI.

    An S3 URI can look like any of the following:

        s3://accesskey:secretkey@s3.amazonaws.com/bucket/key-id
        s3+https://accesskey:secretkey@s3.amazonaws.com/bucket/key-id

    The s3+https:// URIs indicate there is an HTTPS s3service URL
    """
    def process_specs(self):
        self.scheme = self.specs.get('scheme', 's3')
        self.accesskey = self.specs.get('accesskey')
        self.secretkey = self.specs.get('secretkey')
        s3_host = self.specs.get('s3serviceurl')
        self.bucket = self.specs.get('bucket')
        self.key = self.specs.get('key')

        if s3_host.startswith('https://'):
            self.scheme = 's3+https'
            s3_host = s3_host[len('https://'):].strip('/')
        elif s3_host.startswith('http://'):
            s3_host = s3_host[len('http://'):].strip('/')
        self.s3serviceurl = s3_host.strip('/')

    def _get_credstring(self):
        if self.accesskey:
            return '%s:%s@' % (self.accesskey, self.secretkey)
        return ''

    def get_uri(self):
        return "%s://%s%s/%s/%s" % (self.scheme, self._get_credstring(),
                                    self.s3serviceurl, self.bucket, self.key)

    def parse_uri(self, uri):
        """Parse URLs.

        Note that an Amazon AWS secret key can contain the forward slash,
        which is entirely retarded, and breaks urlparse miserably.
        This function works around that issue.
        """
        # Make sure that URIs that contain multiple schemes, such as:
        # s3://accesskey:secretkey@https://s3.amazonaws.com/bucket/key-id
        # are immediately rejected.
        if uri.count('://') != 1:
            reason = ("URI cannot contain more than one occurrence "
                      "of a scheme. If you have specified a URI like "
                      "s3://accesskey:secretkey@"
                      "https://s3.amazonaws.com/bucket/key-id"
                      ", you need to change it to use the "
                      "s3+https:// scheme, like so: "
                      "s3+https://accesskey:secretkey@"
                      "s3.amazonaws.com/bucket/key-id")
            LOG.info("Invalid store uri: %s", reason)
            raise exceptions.BadStoreUri(uri=uri)

        pieces = urllib.parse.urlparse(uri)
        self.validate_schemas(uri, valid_schemas=(
            's3://', 's3+http://', 's3+https://'))
        self.scheme = pieces.scheme
        path = pieces.path.strip('/')
        netloc = pieces.netloc.strip('/')
        entire_path = (netloc + '/' + path).strip('/')

        if '@' in uri:
            creds, path = entire_path.split('@')
            cred_parts = creds.split(':')

            try:
                self.accesskey = cred_parts[0]
                self.secretkey = cred_parts[1]
            except IndexError:
                LOG.error("Badly formed S3 credentials")
                raise exceptions.BadStoreUri(uri=uri)
        else:
            self.accesskey = None
            path = entire_path
        try:
            path_parts = path.split('/')
            self.key = path_parts.pop()
            self.bucket = path_parts.pop()
            if path_parts:
                self.s3serviceurl = '/'.join(path_parts).strip('/')
            else:
                LOG.error("Badly formed S3 URI. Missing s3 service URL.")
                raise exceptions.BadStoreUri(uri=uri)
        except IndexError:
            LOG.error("Badly formed S3 URI")
            raise exceptions.BadStoreUri(uri=uri)


class Store(glance_store.driver.Store):
    """An implementation of the s3 adapter."""

    _CAPABILITIES = capabilities.BitMasks.RW_ACCESS
    OPTIONS = _S3_OPTS
    EXAMPLE_URL = "s3://<ACCESS_KEY>:<SECRET_KEY>@<S3_URL>/<BUCKET>/<OBJ>"

    READ_CHUNKSIZE = 64 * units.Ki
    WRITE_CHUNKSIZE = 5 * units.Mi

    @staticmethod
    def get_schemes():
        return 's3', 's3+http', 's3+https'

    def configure_add(self):
        """
        Configure the Store to use the stored configuration options
        Any store that needs special configuration should implement
        this method. If the store was not able to successfully configure
        itself, it should raise `exceptions.BadStoreConfiguration`
        """
        if boto_session is None:
            reason = _("boto3 or botocore is not available.")
            LOG.error(reason)
            raise exceptions.BadStoreConfiguration(store_name="s3",
                                                   reason=reason)

        self.s3_host = self._option_get('s3_store_host')
        self.region_name = self._option_get('s3_store_region_name')
        self.access_key = self._option_get('s3_store_access_key')
        self.secret_key = self._option_get('s3_store_secret_key')
        self.bucket = self._option_get('s3_store_bucket')

        self.scheme = 's3'
        if self.s3_host.startswith('https://'):
            self.scheme = 's3+https'
            self.full_s3_host = self.s3_host
        elif self.s3_host.startswith('http://'):
            self.full_s3_host = self.s3_host
        else:  # Defaults http
            self.full_s3_host = 'http://' + self.s3_host

        _s3_obj_size = self._option_get('s3_store_large_object_size')
        self.s3_store_large_object_size = _s3_obj_size * units.Mi
        _s3_ck_size = self._option_get('s3_store_large_object_chunk_size')
        _s3_ck_min = DEFAULT_LARGE_OBJECT_MIN_CHUNK_SIZE
        if _s3_ck_size < _s3_ck_min:
            reason = _("s3_store_large_object_chunk_size must be at "
                       "least %d MB.") % _s3_ck_min
            LOG.error(reason)
            raise exceptions.BadStoreConfiguration(store_name="s3",
                                                   reason=reason)
        self.s3_store_large_object_chunk_size = _s3_ck_size * units.Mi

        self.s3_store_thread_pools = self._option_get('s3_store_thread_pools')
        if self.s3_store_thread_pools <= 0:
            reason = _("s3_store_thread_pools must be a positive "
                       "integer. %s") % self.s3_store_thread_pools
            LOG.error(reason)
            raise exceptions.BadStoreConfiguration(store_name="s3",
                                                   reason=reason)

        if self.backend_group:
            self._set_url_prefix()

    def _set_url_prefix(self):
        s3_host = self.s3_host
        if s3_host.startswith('http://'):
            s3_host = s3_host[len('http://'):]
        elif s3_host.startswith('https://'):
            s3_host = s3_host[len('https://'):]

        self._url_prefix = "%s://%s:%s@%s/%s" % (self.scheme, self.access_key,
                                                 self.secret_key, s3_host,
                                                 self.bucket)

    def _option_get(self, param):
        if self.backend_group:
            store_conf = glance_store.driver.BackendGroupConfiguration(
                self.OPTIONS, self.backend_group, conf=self.conf)
        else:
            store_conf = self.conf.glance_store

        result = getattr(store_conf, param)
        if not result:
            if param == 's3_store_create_bucket_on_put':
                return result
            if param == 's3_store_large_object_size':
                return result
            if param == 's3_store_region_name':
                return result
            if param == 's3_store_cacert':
                return result
            if param == 's3_store_enable_data_integrity_protection':
                return result
            if param == 's3_store_request_checksum_calculation':
                return result
            if param == 's3_store_response_checksum_validation':
                return result
            reason = _("Could not find %s in configuration options.") % param
            LOG.error(reason)
            raise exceptions.BadStoreConfiguration(store_name="s3",
                                                   reason=reason)
        return result

    def _create_s3_client(self, loc):
        """Create a client object to use when connecting to S3.

        :param loc: `glance_store.location.Location` object, supplied
                    from glance_store.location.get_location_from_uri()
        :returns: An object with credentials to connect to S3
        """
        s3_host = self._option_get('s3_store_host')
        url_format = self._option_get('s3_store_bucket_url_format')
        calling_format = {'addressing_style': url_format}

        session = boto_session.Session(aws_access_key_id=loc.accesskey,
                                       aws_secret_access_key=loc.secretkey)

        # Determine checksum settings with precedence:
        # 1. If s3_store_enable_data_integrity_protection is False,
        #    use 'when_required' for both (data integrity protection disabled)
        # 2. If s3_store_enable_data_integrity_protection is True,
        #    use the fine-grained option values (default: 'when_required')
        if not self._option_get('s3_store_enable_data_integrity_protection'):
            # Data integrity protection disabled, use when_required for both
            request_checksum = 'when_required'
            response_checksum = 'when_required'
        else:
            # Data integrity protection enabled, use fine-grained options
            request_checksum = self._option_get(
                's3_store_request_checksum_calculation')
            response_checksum = self._option_get(
                's3_store_response_checksum_validation')

        # Try to create Config with checksum parameters (boto3 >= 1.36.0)
        # Fall back to Config without them for older versions
        # NOTE(croelandt): This try/except block should be removed and
        # simplified to always pass the checksum parameters once we require
        # boto3 >= 1.36.0 in requirements.txt
        try:
            config = boto_client.Config(
                s3=calling_format,
                request_checksum_calculation=request_checksum,
                response_checksum_validation=response_checksum,
            )
        except TypeError:
            # boto3 < 1.36.0 doesn't support these parameters
            LOG.warning("boto3 version does not support checksum control "
                        "parameters. Upgrade to boto3 >= 1.36.0 to use "
                        "s3_store_enable_data_integrity_protection and "
                        "fine-grained checksum options.")
            config = boto_client.Config(s3=calling_format)
        location = get_s3_location(s3_host)

        bucket_name = loc.bucket
        if (url_format == 'virtual' and
                not boto_utils.check_dns_name(bucket_name)):
            raise boto_exceptions.InvalidDNSNameError(bucket_name=bucket_name)

        region_name, endpoint_url = None, None
        if self.region_name:
            region_name = self.region_name
            endpoint_url = s3_host
        elif location:
            region_name = location
        else:
            endpoint_url = s3_host

        store_cacert = self._option_get('s3_store_cacert')
        return session.client(
            service_name='s3',
            endpoint_url=endpoint_url,
            region_name=region_name,
            use_ssl=(loc.scheme == 's3+https'),
            verify=None if store_cacert == '' else store_cacert,
            config=config)

    def _operation_set(self, loc):
        """Objects and variables frequently used when operating S3 are
        returned together.

        :param loc: `glance_store.location.Location` object, supplied
                     from glance_store.location.get_location_from_uri()
        "returns: tuple of: (1) S3 client object, (2) Bucket name,
                  (3) Image Object name
        """
        return self._create_s3_client(loc), loc.bucket, loc.key

    @capabilities.check
    def get(self, location, offset=0, chunk_size=None, context=None):
        """
        Takes a `glance_store.location.Location` object that indicates
        where to find the image file, and returns a tuple of generator
        (for reading the image file) and image_size

        :param location: `glance_store.location.Location` object, supplied
                         from glance_store.location.get_location_from_uri()
        :raises: `glance_store.exceptions.NotFound` if image does not exist
        """
        loc = location.store_location
        s3_client, bucket, key = self._operation_set(loc)

        if not self._object_exists(s3_client, bucket, key):
            LOG.warning("Could not find key %(key)s in "
                        "bucket %(bucket)s", {'key': key, 'bucket': bucket})
            raise exceptions.NotFound(image=key)

        key = s3_client.get_object(Bucket=bucket, Key=key)

        LOG.debug("Retrieved image object from S3 using s3_host=%(s3_host)s, "
                  "bucket=%(bucket)s key=%(key)s)",
                  {'s3_host': loc.s3serviceurl, 'bucket': bucket, 'key': key})

        cs = self.READ_CHUNKSIZE

        class ResponseIndexable(glance_store.Indexable):
            def another(self):
                try:
                    return next(self.wrapped)
                except StopIteration:
                    return b''

        return (ResponseIndexable(utils.chunkiter(key['Body'], cs),
                                  key['ContentLength']), key['ContentLength'])

    def get_size(self, location, context=None):
        """
        Takes a `glance_store.location.Location` object that indicates
        where to find the image file and returns the image size

        :param location: `glance_store.location.Location` object, supplied
                         from glance_store.location.get_location_from_uri()
        :raises: `glance_store.exceptions.NotFound` if image does not exist
        :rtype: int
        """
        loc = location.store_location
        s3_client, bucket, key = self._operation_set(loc)

        if not self._object_exists(s3_client, bucket, key):
            LOG.warning("Could not find key %(key)s in "
                        "bucket %(bucket)s", {'key': key, 'bucket': bucket})
            raise exceptions.NotFound(image=key)

        key = s3_client.head_object(Bucket=bucket, Key=key)
        return key['ContentLength']

    @capabilities.check
    def add(self, image_id, image_file, image_size, hashing_algo, context=None,
            verifier=None):
        """
        Stores an image file with supplied identifier to the backend
        storage system and returns a tuple containing information
        about the stored image.

        :param image_id: The opaque image identifier
        :param image_file: The image data to write, as a file-like object
        :param image_size: The size of the image data to write, in bytes
        :param hashing_algo: A hashlib algorithm identifier (string)
        :param context: A context object
        :param verifier: An object used to verify signatures for images

        :returns: tuple of: (1) URL in backing store, (2) bytes written,
                  (3) checksum, (4) multihash value, and (5) a dictionary
                  with storage system specific information
        :raises: `glance_store.exceptions.Duplicate` if the image already
                 exists
        """
        loc = StoreLocation(store_specs={'scheme': self.scheme,
                                         'bucket': self.bucket,
                                         'key': image_id,
                                         's3serviceurl': self.full_s3_host,
                                         'accesskey': self.access_key,
                                         'secretkey': self.secret_key},
                            conf=self.conf,
                            backend_group=self.backend_group)

        s3_client, bucket, key = self._operation_set(loc)

        if not self._bucket_exists(s3_client, bucket):
            if self._option_get('s3_store_create_bucket_on_put'):
                self._create_bucket(s3_client,
                                    self._option_get('s3_store_host'),
                                    bucket,
                                    self._option_get('s3_store_region_name'))
            else:
                msg = (_("The bucket %s does not exist in "
                         "S3. Please set the "
                         "s3_store_create_bucket_on_put option "
                         "to add bucket to S3 automatically.") % bucket)
                raise glance_store.BackendException(msg)

        LOG.debug("Adding image object to S3 using (s3_host=%(s3_host)s, "
                  "bucket=%(bucket)s, key=%(key)s)",
                  {'s3_host': self.s3_host, 'bucket': bucket, 'key': key})

        if not self._object_exists(s3_client, bucket, key):
            if image_size < self.s3_store_large_object_size:
                return self._add_singlepart(s3_client=s3_client,
                                            image_file=image_file,
                                            bucket=bucket,
                                            key=key,
                                            loc=loc,
                                            hashing_algo=hashing_algo,
                                            verifier=verifier,
                                            image_size=image_size)

            return self._add_multipart(s3_client=s3_client,
                                       image_file=image_file,
                                       image_size=image_size,
                                       bucket=bucket,
                                       key=key,
                                       loc=loc,
                                       hashing_algo=hashing_algo,
                                       verifier=verifier)
        LOG.warning("S3 already has an image with bucket ID %(bucket)s, "
                    "key %(key)s", {'bucket': bucket, 'key': key})
        raise exceptions.Duplicate(image=key)

    def _add_singlepart(self, s3_client, image_file, bucket, key, loc,
                        hashing_algo, verifier, image_size):
        """Stores an image file with a single part upload to S3 backend.

        :param s3_client: An object with credentials to connect to S3
        :param image_file: The image data to write, as a file-like object
        :param bucket: S3 bucket name
        :param key: The object name to be stored (image identifier)
        :param loc: `glance_store.location.Location` object, supplied
                    from glance_store.location.get_location_from_uri()
        :param hashing_algo: A hashlib algorithm identifier (string)
        :param verifier: An object used to verify signatures for images
        :param image_size: The size of the image data to write,
                           in bytes
        :returns: tuple of: (1) URL in backing store, (2) bytes written,
                  (3) checksum, (4) multihash value, and (5) a dictionary
                  with storage system specific information
        """
        os_hash_value = utils.get_hasher(hashing_algo, False)
        checksum = utils.get_hasher('md5', False)
        image_data = b''
        total_bytes = 0
        for chunk in utils.chunkreadable(image_file, self.WRITE_CHUNKSIZE):
            total_bytes += len(chunk)
            if image_size != 0 and total_bytes > image_size:
                # Size exceeded, abort upload
                raise glance_store.Invalid(
                    _("Size exceeds: expected %(expected)d bytes, "
                      "got %(actual)d bytes") %
                    {'expected': image_size, 'actual': total_bytes})
            image_data += chunk
            os_hash_value.update(chunk)
            checksum.update(chunk)
            if verifier:
                verifier.update(chunk)

        # Final size check after reading all chunks
        if image_size != 0 and total_bytes != image_size:
            raise glance_store.Invalid(
                _("Size mismatch: expected %(expected)d bytes, "
                  "got %(actual)d bytes") %
                {'expected': image_size, 'actual': total_bytes})

        s3_client.put_object(Body=image_data,
                             Bucket=bucket,
                             Key=key)
        hash_hex = os_hash_value.hexdigest()
        checksum_hex = checksum.hexdigest()

        # Add store backend information to location metadata
        metadata = {}
        if self.backend_group:
            metadata['store'] = self.backend_group

        LOG.debug("Wrote %(size)d bytes to S3 key named %(key)s "
                  "with checksum %(checksum)s",
                  {'size': total_bytes, 'key': key, 'checksum': checksum_hex})

        return loc.get_uri(), total_bytes, checksum_hex, hash_hex, metadata

    def _add_multipart(self, s3_client, image_file, image_size, bucket,
                       key, loc, hashing_algo, verifier):
        """Stores an image file with a multi part upload to S3 backend.

        :param s3_client: An object with credentials to connect to S3
        :param image_file: The image data to write, as a file-like object
        :param bucket: S3 bucket name
        :param key: The object name to be stored (image identifier)
        :param loc: `glance_store.location.Location` object, supplied
                    from glance_store.location.get_location_from_uri()
        :param hashing_algo: A hashlib algorithm identifier (string)
        :param verifier: An object used to verify signatures for images
        :returns: tuple of: (1) URL in backing store, (2) bytes written,
                  (3) checksum, (4) multihash value, and (5) a dictionary
                  with storage system specific information
        """
        os_hash_value = utils.get_hasher(hashing_algo, False)
        checksum = utils.get_hasher('md5', False)
        pool_size = self.s3_store_thread_pools
        # Replace eventlet.GreenPool with ThreadPoolExecutor
        with futures.ThreadPoolExecutor(
                max_workers=pool_size) as executor:
            # Create a list to store the futures
            futures_list = []
            mpu = s3_client.create_multipart_upload(Bucket=bucket, Key=key)
            upload_id = mpu['UploadId']
            LOG.debug("Multipart initiate key=%(key)s, UploadId=%(UploadId)s",
                      {'key': key, 'UploadId': upload_id})
            cstart = 0
            plist = []

            chunk_size = int(math.ceil(float(image_size) / MAX_PART_NUM))
            write_chunk_size = max(self.s3_store_large_object_chunk_size,
                                   chunk_size)
            it = utils.chunkreadable(image_file, self.WRITE_CHUNKSIZE)
            buffered_chunk = b''
            total_bytes = 0
            while True:
                try:
                    buffered_clen = len(buffered_chunk)
                    if buffered_clen < write_chunk_size:
                        # keep reading data
                        read_chunk = next(it)
                        buffered_chunk += read_chunk
                        continue
                    else:
                        write_chunk = buffered_chunk[:write_chunk_size]
                        remained_data = buffered_chunk[write_chunk_size:]
                        total_bytes += len(write_chunk)
                        if image_size != 0 and total_bytes > image_size:
                            # Abort multipart upload immediately
                            s3_client.abort_multipart_upload(
                                Bucket=bucket, Key=key, UploadId=upload_id)
                            raise glance_store.Invalid(
                                _("Size exceeds: expected "
                                  "%(expected)d "
                                  "bytes, got %(actual)d bytes") %
                                {'expected': image_size,
                                 'actual': total_bytes})
                        os_hash_value.update(write_chunk)
                        checksum.update(write_chunk)
                        if verifier:
                            verifier.update(write_chunk)
                        fp = io.BytesIO(write_chunk)
                        fp.seek(0)
                        part = UploadPart(
                            mpu, fp, cstart + 1, len(write_chunk))
                        # Spawn thread to upload part
                        futures_list.append(executor.submit(
                            run_upload, s3_client, bucket, key, part))
                        plist.append(part)
                        cstart += 1
                        buffered_chunk = remained_data
                except StopIteration:
                    if len(buffered_chunk) > 0:
                        # Write the last chunk data
                        write_chunk = buffered_chunk
                        total_bytes += len(write_chunk)
                        os_hash_value.update(write_chunk)
                        checksum.update(write_chunk)
                        if verifier:
                            verifier.update(write_chunk)
                        fp = io.BytesIO(write_chunk)
                        fp.seek(0)
                        part = UploadPart(
                            mpu, fp, cstart + 1, len(write_chunk))
                        futures_list.append(executor.submit(
                            run_upload, s3_client, bucket, key, part))
                        plist.append(part)
                    break

        # Wait for all uploads to finish
        futures.wait(futures_list)

        # Final size validation after all parts uploaded
        if image_size != 0 and total_bytes != image_size:
            # Abort upload if mismatch (redundant check)
            s3_client.abort_multipart_upload(
                Bucket=bucket, Key=key, UploadId=upload_id)
            raise glance_store.Invalid(
                _("Size mismatch: expected %(expected)d "
                  "bytes, got %(actual)d bytes") %
                {'expected': image_size, 'actual': total_bytes})

        # Check success status
        success = all(p.success for p in plist)
        total_size = sum(p.size for p in plist)

        if success:
            hash_hex = os_hash_value.hexdigest()
            checksum_hex = checksum.hexdigest()
            # Add store backend information to location metadata
            metadata = {}
            if self.backend_group:
                metadata['store'] = self.backend_group
            # If no parts were uploaded, this was a zero-byte image. This is
            # the case when a BFV server creates a snapshot resulting in a 0
            # byte image file. Since we cannot complete a multipart upload
            # with zero parts, so we must abort it and create the empty
            # object directly.
            if not plist:
                LOG.info("Not doing multipart upload for zero-byte image "
                         "key=%(key)s, UploadId=%(UploadId)s",
                         {'key': key, 'UploadId': upload_id})
                s3_client.abort_multipart_upload(Bucket=bucket, Key=key,
                                                 UploadId=upload_id)

                LOG.debug("Creating zero-byte object directly using "
                          "singlepart for key=%s", key)
                s3_client.put_object(Body=b'', Bucket=bucket, Key=key)
                LOG.info("Singlepart upload completed. "
                         "Wrote %(total_size)d bytes to S3 key "
                         "named %(key)s "
                         "with checksum %(checksum)s",
                         {'key': key, 'total_size': total_size,
                          'checksum': checksum_hex})
            else:
                # Complete the multipart upload
                pedict = {p.partnum: p.etag[p.partnum] for p in plist}
                mpu_list = self._get_mpu_list(pedict)
                s3_client.complete_multipart_upload(Bucket=bucket,
                                                    Key=key,
                                                    MultipartUpload=mpu_list,
                                                    UploadId=upload_id)

                LOG.info("Multipart complete key=%(key)s "
                         "UploadId=%(UploadId)s "
                         "Wrote %(total_size)d bytes to S3 key "
                         "named %(key)s "
                         "with checksum %(checksum)s",
                         {'key': key, 'UploadId': upload_id,
                          'total_size': total_size, 'checksum': checksum_hex})
            return loc.get_uri(), total_size, checksum_hex, hash_hex, metadata

        # Abort
        s3_client.abort_multipart_upload(Bucket=bucket, Key=key,
                                         UploadId=upload_id)
        LOG.error("Some parts failed to upload to S3. "
                  "Aborted the key=%s", key)
        msg = _("Failed to add image object to S3. key=%s") % key
        raise glance_store.BackendException(msg)

    @capabilities.check
    def delete(self, location, context=None):
        """
        Takes a `glance_store.location.Location` object that indicates
        where to find the image file to delete.

        :param location: `glance_store.location.Location` object, supplied
                         from glance_store.location.get_location_from_uri()

        :raises: NotFound if image does not exist;
                InUseByStore if image is in use or snapshot unprotect failed
        """
        loc = location.store_location
        s3_client, bucket, key = self._operation_set(loc)

        if not self._object_exists(s3_client, bucket, key):
            LOG.warning("Could not find key %(key)s in bucket %(bucket)s",
                        {'key': key, 'bucket': bucket})
            raise exceptions.NotFound(image=key)

        LOG.debug("Deleting image object from S3 using s3_host=%(s3_host)s, "
                  "bucket=%(bucket)s, key=%(key)s)",
                  {'s3_host': loc.s3serviceurl, 'bucket': bucket, 'key': key})

        return s3_client.delete_object(Bucket=bucket, Key=key)

    @staticmethod
    def _bucket_exists(s3_client, bucket):
        """Check whether bucket exists in the S3.

        :param s3_client: An object with credentials to connect to S3
        :param bucket: S3 bucket name
        :returns: boolean value; If the value is true, the bucket is exist
                  if false, it is not.
        :raises: BadStoreConfiguration if cannot connect to S3 successfully
        """
        try:
            s3_client.head_bucket(Bucket=bucket)
        except boto_exceptions.ClientError as e:
            error_code = e.response['Error']['Code']
            if error_code == '404':
                return False
            msg = "Failed to get bucket info: %s" % e
            LOG.error(msg)
            raise glance_store.BadStoreConfiguration(store_name='s3',
                                                     reason=msg)
        else:
            return True

    @staticmethod
    def _object_exists(s3_client, bucket, key):
        """Check whether object exists in the specific bucket of S3.

        :param s3_client: An object with credentials to connect to S3
        :param bucket: S3 bucket name
        :param key: The image object name
        :returns: boolean value; If the value is true, the object is exist
                  if false, it is not.
        :raises: BadStoreConfiguration if cannot connect to S3 successfully
        """
        try:
            s3_client.head_object(Bucket=bucket, Key=key)
        except boto_exceptions.ClientError as e:
            error_code = e.response['Error']['Code']
            if error_code == '404':
                return False
            msg = "Failed to get object info: %s" % e
            LOG.error(msg)
            raise glance_store.BadStoreConfiguration(store_name='s3',
                                                     reason=msg)
        else:
            return True

    @staticmethod
    def _create_bucket(s3_client, s3_host, bucket, region_name=None):
        """Create bucket into the S3.

        :param s3_client: An object with credentials to connect to S3
        :param s3_host: S3 endpoint url
        :param bucket: S3 bucket name
        :param region_name: An optional region_name. If not provided, will try
               to compute it from s3_host
        :raises: BadStoreConfiguration if cannot connect to S3 successfully
        """
        if region_name:
            region = region_name
        else:
            region = get_s3_location(s3_host)
        try:
            s3_client.create_bucket(
                Bucket=bucket,
            ) if region == '' else s3_client.create_bucket(
                Bucket=bucket,
                CreateBucketConfiguration={
                    'LocationConstraint': region
                }
            )
        except boto_exceptions.ClientError as e:
            msg = "Failed to add bucket to S3: %s" % e
            LOG.error(msg)
            raise glance_store.BadStoreConfiguration(store_name='s3',
                                                     reason=msg)

    @staticmethod
    def _get_mpu_list(pedict):
        """Convert an object type and struct for use in
        boto3.client('s3').complete_multipart_upload.

        :param pedict: dict which containing UploadPart.etag
        :returns: list with pedict converted properly
        """
        return {
            'Parts': [
                {
                    'PartNumber': pnum,
                    'ETag': etag
                } for pnum, etag in pedict.items()
            ]
        }


def get_s3_location(s3_host):
    """Get S3 region information from ``s3_store_host``.

    :param s3_host: S3 endpoint url
    :returns: string value; region information which user wants to use on
              Amazon S3, and if user wants to use S3 compatible storage,
              returns ''
    """
    # NOTE(arnaud): maybe get rid of hardcoded amazon stuff here?
    locations = {
        's3.amazonaws.com': '',
        's3-us-east-1.amazonaws.com': 'us-east-1',
        's3-us-east-2.amazonaws.com': 'us-east-2',
        's3-us-west-1.amazonaws.com': 'us-west-1',
        's3-us-west-2.amazonaws.com': 'us-west-2',
        's3-ap-east-1.amazonaws.com': 'ap-east-1',
        's3-ap-south-1.amazonaws.com': 'ap-south-1',
        's3-ap-northeast-1.amazonaws.com': 'ap-northeast-1',
        's3-ap-northeast-2.amazonaws.com': 'ap-northeast-2',
        's3-ap-northeast-3.amazonaws.com': 'ap-northeast-3',
        's3-ap-southeast-1.amazonaws.com': 'ap-southeast-1',
        's3-ap-southeast-2.amazonaws.com': 'ap-southeast-2',
        's3-ca-central-1.amazonaws.com': 'ca-central-1',
        's3-cn-north-1.amazonaws.com.cn': 'cn-north-1',
        's3-cn-northwest-1.amazonaws.com.cn': 'cn-northwest-1',
        's3-eu-central-1.amazonaws.com': 'eu-central-1',
        's3-eu-west-1.amazonaws.com': 'eu-west-1',
        's3-eu-west-2.amazonaws.com': 'eu-west-2',
        's3-eu-west-3.amazonaws.com': 'eu-west-3',
        's3-eu-north-1.amazonaws.com': 'eu-north-1',
        's3-sa-east-1.amazonaws.com': 'sa-east-1'
    }
    # strip off scheme and port if present
    key = re.sub(r'^(https?://)?(?P<host>[^:]+[^/])(:[0-9]+)?/?$',
                 r'\g<host>',
                 s3_host)
    return locations.get(key, '')
